"""Neural Network classes and helper functions for PyTorch."""
from abc import ABC
import logging
from typing import Any, Iterable, List, Optional, Tuple, Type, Union, cast

import torch
import torch.nn as nn
import torchvision.models as tv_models

from bitfount.backends.pytorch.models.torch_functions.mish import Mish
from bitfount.models.nn import _ConvNeuralNet, _FeedForwardNeuralNet, _NeuralNet

logger = logging.getLogger(__name__)

TORCHVISION_CLASSIFICATION_MODELS = {
    name: func
    for name, func in vars(tv_models).items()
    if callable(func) and not isinstance(func, type)
}


class _PyTorchNeuralNet(nn.Module, _NeuralNet, ABC):
    """Base abstract class for all PyTorch-implemented neural networks."""

    pass


class _PyTorchConvNeuralNet(_PyTorchNeuralNet, _ConvNeuralNet):
    """Simple convolutional neural network architecture in PyTorch."""

    def __init__(
        self,
        layer_sizes: Iterable[Tuple[int, int]],
        dropout_probs: List[float],
        mish: bool,
        head_sizes: Iterable[Tuple[int, int]],
        ff_layer_sizes: Iterable[Tuple[int, int]],
        ff_dropout_probs: List[float],
        kernel_size: int,
        padding: int,
        stride: int,
        pooling_function: str,
    ):
        super().__init__()
        # CONVOLUTIONAL BLOCKS
        self.layers = nn.ModuleList(
            [
                nn.Conv2d(
                    in_, out_, kernel_size=kernel_size, stride=stride, padding=padding
                )
                for in_, out_ in layer_sizes
            ]
        )
        activation_function = Mish if mish else nn.ReLU
        self.activations = nn.ModuleList([activation_function() for _ in layer_sizes])
        self.batch_norms = nn.ModuleList(
            [nn.BatchNorm2d(size) for _, size in layer_sizes]
        )
        self.dropouts = nn.ModuleList([nn.Dropout2d(i) for i in dropout_probs])
        pooling_module: Union[Type[nn.AvgPool2d], Type[nn.MaxPool2d]]
        if pooling_function == "max":
            pooling_module = nn.MaxPool2d
        elif pooling_function == "avg":
            pooling_module = nn.AvgPool2d

        self.pooling_functions = nn.ModuleList(
            [pooling_module(kernel_size=2) for _ in layer_sizes]
        )

        # FEEDFORWARD LAYERS
        self.ff_layers = nn.ModuleList(
            [nn.Linear(in_, out_) for in_, out_ in ff_layer_sizes]
        )
        self.ff_activations = nn.ModuleList(
            [activation_function() for _ in ff_layer_sizes]
        )
        self.ff_batch_norms = nn.ModuleList(
            [nn.BatchNorm1d(size) for _, size in ff_layer_sizes]
        )
        self.ff_dropouts = nn.ModuleList([nn.Dropout(i) for i in ff_dropout_probs])

        # OUTPUT LAYER
        self.heads = nn.ModuleList([nn.Linear(in_, out_) for in_, out_ in head_sizes])

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Forward pass in model."""
        for layer, activation, batchnorm, dropout, pooling_function in zip(
            self.layers,
            self.activations,
            self.batch_norms,
            self.dropouts,
            self.pooling_functions,
        ):
            x = layer(x)
            x = activation(x)
            x = batchnorm(x)
            x = dropout(x)
            x = pooling_function(x)

        x = x.view(x.size(0), -1)

        for layer, activation, batchnorm, dropout in zip(
            self.ff_layers, self.ff_activations, self.ff_batch_norms, self.ff_dropouts
        ):
            x = layer(x)
            x = activation(x)
            x = batchnorm(x)
            x = dropout(x)

        return tuple(head(x) for head in self.heads)


class _PyTorchFeedForwardNeuralNet(_PyTorchNeuralNet, _FeedForwardNeuralNet):
    """Simple feedforward neural network architecture in PyTorch."""

    def __init__(
        self,
        embedding_sizes: Iterable[Tuple[int, int]],
        emb_dropout: float,
        n_cont: int,
        layer_sizes: Iterable[Tuple[int, int]],
        dropout_probs: List[float],
        mish: bool,
        head_sizes: Iterable[Tuple[int, int]],
    ):
        super().__init__()

        # INPUT LAYER
        self.embeddings = nn.ModuleList(
            [nn.Embedding(categories, size) for categories, size in embedding_sizes]
        )
        self.emb_drop = nn.Dropout(emb_dropout)
        self.bn_cont = nn.BatchNorm1d(n_cont)
        self.n_cont = n_cont
        # HIDDEN LAYERS
        self.layers = nn.ModuleList([nn.Linear(in_, out_) for in_, out_ in layer_sizes])
        activation_function = Mish if mish else nn.ReLU
        self.activations = nn.ModuleList([activation_function() for _ in layer_sizes])
        self.batch_norms = nn.ModuleList(
            [nn.BatchNorm1d(size) for _, size in layer_sizes]
        )
        self.dropouts = nn.ModuleList([nn.Dropout(i) for i in dropout_probs])

        # OUTPUT LAYER
        self.heads = nn.ModuleList([nn.Linear(in_, out_) for in_, out_ in head_sizes])

    def forward(self, x: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, ...]:
        """Forward pass in model."""
        x_cat, x_cont = x
        if x_cont.nelement() != 0 and x_cat.nelement() != 0:
            fwd_pass_cat = [
                cast(torch.Tensor, embedding(x_cat[idx]))
                for idx, embedding in enumerate(self.embeddings)
            ]
            fwd_pass = torch.cat(fwd_pass_cat, 1)
            fwd_pass = self.emb_drop(fwd_pass)
            fwd_pass2 = self.bn_cont(x_cont)
            fwd_pass = torch.cat([fwd_pass, fwd_pass2], 1)
        elif x_cont.nelement() != 0:
            fwd_pass = self.bn_cont(x_cont)
        elif x_cat.nelement() != 0:
            fwd_pass_cat = [
                cast(torch.Tensor, embedding(x_cat[idx]))
                for idx, embedding in enumerate(self.embeddings)
            ]
            fwd_pass = torch.cat(fwd_pass_cat, 1)
            fwd_pass = self.emb_drop(fwd_pass)

        for layer, activation, batchnorm, dropout in zip(
            self.layers, self.activations, self.batch_norms, self.dropouts
        ):
            fwd_pass = layer(fwd_pass)
            fwd_pass = activation(fwd_pass)
            fwd_pass = batchnorm(fwd_pass)
            fwd_pass = dropout(fwd_pass)

        return tuple(head(fwd_pass) for head in self.heads)


class _PyTorchLogisticRegression(_PyTorchNeuralNet, _FeedForwardNeuralNet):
    def __init__(
        self,
        num_classes: int,
        input_dim: Optional[int] = None,
        num_continuous: Optional[int] = None,
        embedding_sizes: Optional[Iterable[Tuple[int, int]]] = None,
        embedding_dropout_frac: float = 0.04,
        bias: bool = True,
    ):
        """Create a new logistic regression model.

        Input size can either be specified directly (input_dim) or can be
        automatically calculated if using categorical embeddings by providing
        num_continuous and embedding_sizes.

        Args:
            num_classes: Number of classes in the target feature.
            input_dim: The size of the input vector to the model. Cannot be provided
                if num_continuous or embedding_sizes are.
            num_continuous: The number of continuous features in the input.
            embedding_sizes: The sizes of the categorical embeddings for the
                categorical features in the input.
            embedding_dropout_frac: The dropout probability to apply to categorical
                variables if using embed_categorical.
            bias: Whether the underlying linear layer should learn an additive bias.
                Default True.
        """
        super().__init__()

        # If input dimensions are specified, ignore any of the embedding settings
        if input_dim and any(i is not None for i in (num_continuous, embedding_sizes)):
            logger.warning(
                "Cannot specify both input_dim and either of num_continuous or"
                " embedding_sizes. Will default to using input_dim."
            )
            num_continuous = None
            embedding_sizes = None

        # If neither input_dim nor embedding settings are provided, throw error
        if input_dim is None and embedding_sizes is None:
            raise ValueError("One of input_dim or embedding_sizes must be provided.")

        # If embedding_sizes is provided, num_continuous must be, and vice-versa
        if (embedding_sizes is None) != (num_continuous is None):
            raise ValueError(
                "If one of embedding_sizes or num_continuous is provided, both must be"
            )

        self.num_classes = num_classes
        self.input_dim = input_dim
        self.num_continuous = num_continuous
        self.embedding_sizes = embedding_sizes
        self.embedding_dropout_frac = embedding_dropout_frac
        self.bias = bias

        # Calculate input dim if not explicitly provided
        if self.input_dim is None:
            # Above checks guarantee that if input_dim is None, these must be set
            assert self.num_continuous is not None  # nosec[assert_used]
            assert self.embedding_sizes is not None  # nosec[assert_used]
            # Input dim is equal to the number of continuous elements in the
            # input vector and the size of the embeddings for the categorical elements.
            self.input_dim = self.num_continuous + sum(
                size for _, size in self.embedding_sizes
            )

        # INPUT EMBEDDING LAYERS
        if self.embedding_sizes:
            self.category_embeddings: Optional[nn.ModuleList] = nn.ModuleList(
                [
                    nn.Embedding(categories, size)
                    for categories, size in self.embedding_sizes
                ]
            )
            self.category_embeddings_dropout: Optional[nn.Dropout] = nn.Dropout(
                self.embedding_dropout_frac
            )
        else:
            self.category_embeddings = None
            self.category_embeddings_dropout = None

        # MAIN LAYER
        self.linear = nn.Linear(
            in_features=self.input_dim, out_features=self.num_classes, bias=self.bias
        )

    def forward(self, x: Tuple[torch.Tensor, torch.Tensor]) -> torch.Tensor:
        # Split into continuous and categorical features
        x_categorical: torch.Tensor
        x_continuous: torch.Tensor
        x_categorical, x_continuous = x

        to_concat: List[torch.Tensor] = []

        # Handle categorical features, if any
        if x_categorical.nelement() != 0:
            if self.category_embeddings:
                # If category_embeddings is set, category_embeddings_dropout
                # must be too
                assert (  # nosec[assert_used]
                    self.category_embeddings_dropout is not None
                )

                # Embed categorical if required
                fwd_pass_categorical = torch.cat(
                    [
                        embedding(x_categorical[idx])
                        for idx, embedding in enumerate(self.category_embeddings)
                    ],
                    dim=1,
                )
                fwd_pass_categorical = self.category_embeddings_dropout(
                    fwd_pass_categorical
                )
            else:
                # Otherwise, use as is
                fwd_pass_categorical = x_categorical

            to_concat.append(fwd_pass_categorical)

        # Add in continuous features if present
        if x_continuous.nelement() != 0:
            to_concat.append(x_continuous)

        # We return the unactivated output vector to allow it to be used
        # in various settings.
        input_vector = torch.cat(to_concat, dim=1)

        # Linear expects a FloatTensor, so if it's not (i.e. because we only have
        # non-embedded categorical features) we must treat it as such.
        if not torch.is_floating_point(input_vector):
            input_vector = input_vector.float()

        output_vector: torch.Tensor = self.linear(input_vector)
        return output_vector


def _get_torchvision_classification_model(
    model_name: str, pretrained: bool, num_classes: int, **kwargs: Any
) -> nn.Module:
    """Returns a pre-existing torchvision model.

    This function returns the torchvision classification model corresponding to
    `model_name`. Importantly, it resizes the final layer to make it appropriate
    for the task. Since this is different for every model, this must be hard-coded

    Adapted from pytorch docs/tutorials
    """
    # Convert model name for consistency
    model_name = model_name.lower()

    if "resnet" in model_name:
        model = TORCHVISION_CLASSIFICATION_MODELS[model_name](
            pretrained=pretrained, **kwargs
        )
        num_ftrs = model.fc.in_features
        model.fc = nn.Linear(num_ftrs, num_classes)
    elif ("alexnet" in model_name) or ("vgg" in model_name):
        model = TORCHVISION_CLASSIFICATION_MODELS[model_name](
            pretrained=pretrained, **kwargs
        )
        num_ftrs = model.classifier[6].in_features
        model.classifier[6] = nn.Linear(num_ftrs, num_classes)
    elif "squeezenet" in model_name:
        model = TORCHVISION_CLASSIFICATION_MODELS[model_name](
            pretrained=pretrained, **kwargs
        )
        model.classifier[1] = nn.Conv2d(
            512, num_classes, kernel_size=(1, 1), stride=(1, 1)
        )
        model.num_classes = num_classes
    elif "densenet" in model_name:
        model = TORCHVISION_CLASSIFICATION_MODELS[model_name](
            pretrained=pretrained, **kwargs
        )
        num_ftrs = model.classifier.in_features
        model.classifier = nn.Linear(num_ftrs, num_classes)
    elif model_name in TORCHVISION_CLASSIFICATION_MODELS:
        raise ValueError("Model reshaping not implemented yet. Choose another model.")
    else:
        raise ValueError("Model name not recognised")

    return cast(nn.Module, model)



"""Testcases for in bitfount/federated/utils.py."""
import pytest

from bitfount.federated.utils import _keys_exist
from tests.utils.helper import unit_test


@unit_test
def test_keys_exist_works() -> None:
    """Tests that _keys_exists works as expected with correct input."""
    data = {
        "spam": {
            "egg": {
                "bacon": "Is good..",
                "sausages": "Spam egg sausages and spam",
                "spam": "does not have much spam in it",
            }
        }
    }
    assert _keys_exist(data, "spam") is True
    assert _keys_exist(data, "spam", "bacon") is False
    assert _keys_exist(data, "spam", "egg") is True
    assert _keys_exist(data, "spam", "egg", "bacon") is True


@unit_test
def test_keys_exist_error_dict() -> None:
    """Tests that _keys_exists works as expected with wrong dict input."""
    data = "spam"
    with pytest.raises(AttributeError):
        _keys_exist(data, "spam")  # type: ignore[arg-type] # Reason: we test wrong input here. # noqa: B950


@unit_test
def test_keys_exist_missing_keys() -> None:
    """Tests that _keys_exists works as expected with wrong keys input."""
    data = {
        "spam": {
            "egg": {
                "bacon": "Is good..",
                "sausages": "Spam egg sausages and spam",
                "spam": "does not have much spam in it",
            }
        }
    }
    with pytest.raises(AttributeError):
        _keys_exist(data)



"""Tests hub helper.py."""
import logging
from pathlib import Path
from typing import Dict, Optional
from unittest.mock import MagicMock, Mock, create_autospec

from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
import pytest
from pytest import fixture
from pytest_lazyfixture import lazy_fixture
from pytest_mock import MockerFixture
import yaml

import bitfount
from bitfount.data.schema import BitfountSchema
from bitfount.federated.encryption import _RSAEncryption
from bitfount.hub.api import BitfountAM, BitfountHub
from bitfount.hub.authentication_flow import (
    _DEVELOPMENT_CLIENT_ID,
    _PRODUCTION_CLIENT_ID,
    _STAGING_CLIENT_ID,
    BitfountSession,
)
from bitfount.hub.exceptions import PodDoesNotExistError
from bitfount.hub.helper import (
    _check_known_pods,
    _create_access_manager,
    _create_bitfount_session,
    _create_bitfounthub,
    _default_bitfounthub,
    _get_pod_public_key,
    _save_key_to_key_store,
    get_pod_schema,
)
from bitfount.hub.types import (
    _DEV_AM_URL,
    _DEV_HUB_URL,
    _STAGING_AM_URL,
    _STAGING_HUB_URL,
    PRODUCTION_AM_URL,
    PRODUCTION_HUB_URL,
)
from tests.bitfount import TEST_SECURITY_FILES
from tests.utils.helper import get_info_logs, unit_test


def _gen_new_public_key() -> RSAPublicKey:
    """Creates a new RSA public key."""
    return _RSAEncryption.generate_key_pair()[1]


def _key_to_str(key: RSAPublicKey) -> str:
    """Converts RSA public key to string."""
    return _RSAEncryption.serialize_public_key(key).decode()


@unit_test
class TestHelperFunctions:
    """Test hub helper functions."""

    @fixture
    def username(self) -> str:
        """Username for tests."""
        return "test_username"

    @fixture
    def public_key_path(self) -> Path:
        """Path to test public key file."""
        return TEST_SECURITY_FILES / "test_public.testkey"

    @fixture
    def mock_key_store(self, monkeypatch: MonkeyPatch, tmp_path: Path) -> Path:
        """A mock keystore for use with modeller."""
        key_store_path = tmp_path / "known_workers.yml"
        monkeypatch.setattr(bitfount.hub.helper, "BITFOUNT_KEY_STORE", key_store_path)
        return key_store_path.expanduser()

    @fixture
    def mock_input(self, monkeypatch: MonkeyPatch) -> MagicMock:
        """Mock `builtins.input`."""
        my_mock = MagicMock()
        monkeypatch.setattr("builtins.input", my_mock)
        return my_mock

    @fixture
    def test_url(self) -> str:
        """Test URL."""
        return "not.a.real.url.com"

    @pytest.mark.parametrize(
        "input_url, expected_client_id",
        [
            (PRODUCTION_HUB_URL, _PRODUCTION_CLIENT_ID),
            (_STAGING_HUB_URL, _STAGING_CLIENT_ID),
            ("localhost:8888", _DEVELOPMENT_CLIENT_ID),
        ],
    )
    def test_create_bitfount_session_with_different_urls(
        self, expected_client_id: str, input_url: str, username: str
    ) -> None:
        """Tests private _create_bitfount_session function with different urls."""
        session = _create_bitfount_session(url=input_url, username=username)
        assert isinstance(session, BitfountSession)
        assert session._client_id == expected_client_id
        assert session.user_storage_path.stem == username

    @pytest.mark.parametrize(
        "environment, input_url, expected_url",
        [
            ("production", None, PRODUCTION_HUB_URL),
            ("staging", None, _STAGING_HUB_URL),
            ("dev", None, _DEV_HUB_URL),
            (None, lazy_fixture("test_url"), lazy_fixture("test_url")),
        ],
        indirect=["environment"],
    )
    def test_create_bitfounthub_with_different_environments(
        self,
        environment: None,
        expected_url: str,
        input_url: Optional[str],
        mock_bitfount_session: Mock,
        monkeypatch: MonkeyPatch,
        username: str,
    ) -> None:
        """Tests create_bitfounthub with production, staging and local urls.

        We mock BitfountSession to avoid authenticating.
        """
        hub = _create_bitfounthub(username=username, url=input_url)
        assert isinstance(hub, BitfountHub)
        assert hub.url == expected_url

    @pytest.mark.parametrize(
        "environment, input_url, expected_url",
        [
            ("production", None, PRODUCTION_AM_URL),
            ("staging", None, _STAGING_AM_URL),
            ("dev", None, _DEV_AM_URL),
            (None, lazy_fixture("test_url"), lazy_fixture("test_url")),
        ],
        indirect=["environment"],
    )
    def test_create_access_manager_with_different_environments(
        self,
        environment: None,
        expected_url: str,
        input_url: Optional[str],
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Tests create_access_manager with production, staging and local urls."""
        session = Mock()

        am = _create_access_manager(session, input_url)
        assert isinstance(am, BitfountAM)
        assert am.access_manager_url == expected_url
        assert am.session == session

    def test__get_pod_public_key(
        self,
        caplog: LogCaptureFixture,
        mock_key_store: Path,
        mocker: MockerFixture,
        public_key_path: Path,
    ) -> None:
        """Tests _get_pod_public_key helper function."""
        # Mock out the hub and RSA key loading components
        mock_load_public_key = mocker.patch.object(_RSAEncryption, "load_public_key")
        mock_public_key: str = "this-is-a-pod-public-key"
        hub = create_autospec(BitfountHub, instance=True)
        hub.get_pod_key.return_value = mock_public_key

        # Mock out the _check_known_pods() function
        mock_check_known_pods = mocker.patch(
            "bitfount.hub.helper._check_known_pods",
            autospec=True,
            side_effect=lambda pod_id, key, store_path: key,
        )

        mock_mkdir = mocker.patch("bitfount.hub.helper.Path.mkdir")

        # BITFOUNTHUB KEY
        # Correct return value
        public_key = _get_pod_public_key("blah/worker_1", hub)
        mock_load_public_key.assert_called_once_with(mock_public_key.encode())
        mock_check_known_pods.assert_called_once()
        mock_mkdir.assert_called_once()
        assert public_key == mock_load_public_key()

        # KEY FROM FILE
        # Key already exists in file, returns it
        mock_load_public_key.reset_mock()
        mock_check_known_pods.reset_mock()
        # Mock out the scenario of _check_known_pods() returning the existing key
        mock_check_known_pods.side_effect = lambda _pod_id, key, _mock_key_store: key
        public_key = _get_pod_public_key(
            "blah/worker_2", hub, {"blah/worker_2": public_key_path}
        )
        mock_load_public_key.assert_called_once_with(public_key_path)
        mock_check_known_pods.assert_called_once_with(
            "blah/worker_2", mock_load_public_key(), mock_key_store
        )
        assert public_key == mock_load_public_key()
        # mkdir shouldn't be called on second call to _get_pod_public_key
        mock_mkdir.assert_called_once()

        # Key files exist, but not for specific user, retrieves from hub
        with caplog.at_level(logging.DEBUG):
            hub.get_pod_key.return_value = mock_public_key
            mock_load_public_key.reset_mock()
            public_key = _get_pod_public_key(
                "blah/worker_3", hub, {"blah/worker_2": public_key_path}
            )
            mock_load_public_key.assert_called_once_with(mock_public_key.encode())
            assert public_key == mock_load_public_key()
            assert "No existing public key file for blah/worker_3" in caplog.text
            # mkdir shouldn't be called on third call to _get_pod_public_key
            mock_mkdir.assert_called_once()

    def test__check_known_pods(
        self, mock_input: MagicMock, mock_key_store: Path
    ) -> None:
        """Tests _check_known_pods helper function.

        Checks that various states (no key exists, key already exists, wrong
        input detected) all get the correct key and that the key store is updated
        each time.
        """
        orig_key: RSAPublicKey = _gen_new_public_key()
        known_workers: Dict[str, str]
        worker_name = "test-worker"
        mock_key_store.parent.mkdir(parents=True, exist_ok=True)
        mock_key_store.touch()

        # No key in yaml, should not require `input()` call
        new_key: RSAPublicKey = _check_known_pods(worker_name, orig_key, mock_key_store)
        mock_input.assert_not_called()
        assert new_key == orig_key
        with open(mock_key_store, "r") as known_workers_file:
            known_workers = yaml.safe_load(known_workers_file)
        assert known_workers[worker_name] == _key_to_str(orig_key)

        # Accept new key, input should be "Y"
        mock_input.reset_mock(return_value=True, side_effect=True)
        mock_input.return_value = "Y"
        diff_key: RSAPublicKey = _gen_new_public_key()
        new_key = _check_known_pods(worker_name, diff_key, mock_key_store)
        mock_input.assert_called_once()
        assert new_key == diff_key
        with open(mock_key_store, "r") as known_workers_file:
            known_workers = yaml.safe_load(known_workers_file)
        assert known_workers[worker_name] == _key_to_str(diff_key)

        # Wrong input (not "Y" or "N") then reject ("N") new key
        mock_input.reset_mock(return_value=True, side_effect=True)
        mock_input.side_effect = ["INCORRECT_INPUT", "N"]
        key_to_reject: RSAPublicKey = _gen_new_public_key()
        new_key = _check_known_pods(worker_name, key_to_reject, mock_key_store)
        assert mock_input.call_count == 2
        assert new_key != key_to_reject
        # have to do str compare here as direct key compare relies on python id(),
        # but we've actually reloaded the key from file.
        assert _key_to_str(new_key) == _key_to_str(diff_key)  # i.e. hasn't changed
        with open(mock_key_store, "r") as known_workers_file:
            known_workers = yaml.safe_load(known_workers_file)
        assert known_workers[worker_name] != _key_to_str(key_to_reject)
        assert (
            known_workers[worker_name] == _key_to_str(new_key) == _key_to_str(diff_key)
        )  # i.e. hasn't changed

        # Return to original key, input should be "Y"
        mock_input.reset_mock(return_value=True, side_effect=True)
        mock_input.return_value = "Y"
        new_key = _check_known_pods(worker_name, orig_key, mock_key_store)
        mock_input.assert_called_once()
        assert new_key == orig_key
        with open(mock_key_store, "r") as known_workers_file:
            known_workers = yaml.safe_load(known_workers_file)
        assert known_workers[worker_name] == _key_to_str(orig_key)

    def test__check_known_pods_returns_if_key_match(
        self, caplog: LogCaptureFixture, mock_key_store: Path, mocker: MockerFixture
    ) -> None:
        """Tests _check_known_pods doesn't re-save an already existing key."""
        public_key = _gen_new_public_key()
        pod_id = "this-is-a-pod-id"

        # Pre-save the key to the key store
        mock_key_store.parent.mkdir(parents=True, exist_ok=True)
        mock_key_store.touch()
        _save_key_to_key_store(mock_key_store, pod_id, _key_to_str(public_key))

        # Mock out _save_key_to_key_store
        mock_save_keys_to_key_store = mocker.patch(
            "bitfount.hub.helper._save_key_to_key_store"
        )

        with caplog.at_level(logging.INFO):
            returned_public_key = _check_known_pods(pod_id, public_key, mock_key_store)

        # Check same key and saving not called unnecessarily
        # We use an `is` check here because it should return the _exact_ same key
        assert returned_public_key is public_key
        mock_save_keys_to_key_store.assert_not_called()
        assert f"Found public key for {pod_id} in key store." in get_info_logs(caplog)

    def test_default_bitfounthub(self, mocker: MockerFixture) -> None:
        """Tests default bitfounthub calls create only if not None."""
        mock = Mock()
        mocker.patch("bitfount.hub.helper._create_bitfounthub", mock)
        _default_bitfounthub(hub=Mock())
        mock.assert_not_called()
        _default_bitfounthub()
        mock.assert_called_once()

    def test_get_pod_schema(self, mocker: MockerFixture) -> None:
        """Test get_pod_schema downloads schema.

        Tests it with save file specified.
        """
        pod_identifier = "fake/pod"
        save_file_path = "save_file/path.txt"

        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_create_hub = mocker.patch(
            "bitfount.hub.helper._create_bitfounthub", return_value=mock_hub
        )

        # Mock out schema download
        mock_schema = create_autospec(BitfountSchema, instance=True)
        mock_hub.get_pod_schema.return_value = mock_schema

        schema = get_pod_schema(pod_identifier, save_file_path=save_file_path)

        # Check hub was created
        mock_create_hub.assert_called_once()

        # Check schema download called
        mock_hub.get_pod_schema.assert_called_once_with(pod_identifier)

        # Check file saved out
        mock_schema.dump.assert_called_once_with(Path(save_file_path))

        # Check return
        assert schema == mock_schema

    def test_get_pod_schema_with_name_only(self, mocker: MockerFixture) -> None:
        """Test get_pod_schema downloads schema.

        Tests it with save file specified.
        """
        username = "username"
        pod_name = "pod_name"

        # Mock out hub creation, set username
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.username = username
        mock_create_hub = mocker.patch(
            "bitfount.hub.helper._create_bitfounthub", return_value=mock_hub
        )

        # Mock out schema download
        mock_schema = create_autospec(BitfountSchema, instance=True)
        mock_hub.get_pod_schema.return_value = mock_schema

        # Call get_pod_schema with name only
        schema = get_pod_schema(pod_name)

        # Check hub was created
        mock_create_hub.assert_called_once()

        # Check schema download called with constructed pod_identifier
        mock_hub.get_pod_schema.assert_called_once_with(f"{username}/{pod_name}")

        # Check return
        assert schema == mock_schema

    def test_get_pod_schema_with_hub_and_username(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test get_pod_schema with hub and username provided."""
        username = "username"
        hub_username = "hub_username"

        # Mock out hub creation, set username
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.username = hub_username

        # Mock out schema download
        mock_schema = create_autospec(BitfountSchema, instance=True)
        mock_hub.get_pod_schema.return_value = mock_schema

        # Call get_pod_schema
        get_pod_schema("pod_name", hub=mock_hub, username=username)

        # Check that provided username is ignored
        assert "Ignoring username argument as hub was provided." in [
            i.message for i in caplog.records
        ]
        assert mock_hub.username != username

    def test__save_keys_to_key_store(
        self, mock_key_store: Path, mocker: MockerFixture
    ) -> None:
        """Tests saving keys to key store."""
        pod_identifier = "fake/pod"
        orig_key: RSAPublicKey = _RSAEncryption.generate_key_pair()[1]
        serialized_key = _RSAEncryption.serialize_public_key(orig_key).decode()
        mock_open = mocker.mock_open(read_data="")
        mocker.patch("builtins.open", mock_open)

        _save_key_to_key_store(mock_key_store, pod_identifier, serialized_key)

        assert mock_open.call_count == 2
        mock_open.assert_any_call(mock_key_store, "r")
        mock_open.assert_any_call(mock_key_store, "w")

    def test__error_thrown_no_pod_key(
        self,
        mocker: MockerFixture,
    ) -> None:
        """Tests _get_pod_public_key helper function."""
        # Mock out the hub and RSA key loading components
        hub = create_autospec(BitfountHub, instance=True)
        hub.get_pod_key.return_value = None
        pod_id = "pod_id"
        err_msg = f"No public key found for pod: {pod_id}"

        # Mock out the _save_keys_to_key_store() function
        mock_save_keys_to_key_store = mocker.patch(
            "bitfount.hub.helper._save_key_to_key_store", autospec=True
        )

        with pytest.raises(PodDoesNotExistError) as error:
            _get_pod_public_key(pod_id, hub)
            assert str(error.value) == err_msg

        assert mock_save_keys_to_key_store.call_count == 0



"""Utility functions concerning data sources."""

from typing import Dict, Union

import numpy as np
import pandas as pd

from bitfount.data.types import DataPathModifiers


def _modify_column(
    column: Union[np.ndarray, pd.Series],
    modifier_dict: DataPathModifiers,
) -> Union[np.ndarray, pd.Series]:
    """Modify the given column.

    Args:
        column: The column you are operating on.
        modifier_dict: A dictionary with the key as the
            prefix/suffix and the value to be prefixed/suffixed.
    """
    # Get the modifier dictionary:
    for modifier_type, modifier_string in modifier_dict.items():
        if modifier_type == "prefix":
            column = modifier_string + column.astype(str)

        elif modifier_type == "suffix":
            column = column.astype(str) + modifier_string
    return column


def _modify_file_paths(
    data: pd.DataFrame, modifiers: Dict[str, DataPathModifiers]
) -> None:
    """Modifies image file paths if provided.

    Args:
        data: The dataframe to modify.
        modifiers: A dictionary with the column name and
            prefix and/or suffix to modify file path.
    """
    for column_name in modifiers:
        # Get the modifier dictionary:
        modifier_dict = modifiers[column_name]
        data[column_name] = _modify_column(data[column_name], modifier_dict)



"""Handling of OIDC challenges."""
from __future__ import annotations

import asyncio
from asyncio import Task, create_task
import base64
from collections import defaultdict
from datetime import datetime, timedelta, timezone
import hashlib
import secrets
from typing import DefaultDict, Dict, Final, Iterator, List
from urllib.parse import urlencode
import webbrowser

from aiohttp import web
from aiohttp.web import Application, AppRunner
from aiohttp.web import Request as AioRequest
from aiohttp.web import Response as AioResponse
from aiohttp.web import TCPSite
import requests
from requests import HTTPError

from bitfount import web_utils
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.transport.identity_verification import (
    _BITFOUNT_MODELLER_PORT,
    _PORT_WAIT_TIMEOUT,
)
from bitfount.federated.transport.identity_verification.types import (
    _HasWebServer,
    _ResponseHandler,
)
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.types import (
    _DeviceCodeDetailsPair,
    _ModellerDeviceCodeDetails,
    _OIDCAuthEndpointResponse,
    _OIDCAuthFlowResponse,
    _OIDCClientID,
    _PodDeviceCodeDetails,
)
from bitfount.hub.authentication_flow import (
    _PRODUCTION_AUTH_DOMAIN,
    AuthEnvironmentError,
    _get_auth_environment,
)
from bitfount.hub.types import _DeviceCodeRequestDict, _DeviceCodeResponseJSON
from bitfount.types import _JSONDict

logger = _get_federated_logger(__name__)

# This forces `requests` to make IPv4 connections
# TODO: [BIT-1443] Remove this once Hub/AM support IPv6
requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950

_OIDC_IDENTITY_VERIFICATION_SCOPES: Final[str] = "execute:task"
_OIDC_IDENTITY_VERIFICATION_AUDIENCE: Final[
    str
] = "https://am.bitfount.com/api/access/modellerProofOfIdentity"

_CALLBACK_ROUTE: Final[str] = "/api/auth/callback/auth0"


class _OIDCWebEndpoint:
    """Web endpoint for OIDC Authorization Code Flow callbacks."""

    def __init__(self) -> None:
        self.all_done = asyncio.Event()
        self._responses: Dict[str, _OIDCAuthEndpointResponse] = {}
        self._initialised = False

        self._id_to_state: Dict[str, str]
        self._state_to_id: Dict[str, str]
        self._urls_iter: Iterator[str]

    def initialise(self, urls: List[str], states: Dict[str, str]) -> None:
        """Initialise the endpoint with the authorize URLs and states to use.

        This has to be done post-init as the webserver takes a while to spin up
        so we want to start it early and provide the data when we are able to.
        """
        self._id_to_state = states
        self._state_to_id = {v: k for k, v in states.items()}

        self._urls_iter = iter(urls)

        self._initialised = True

    def start_processing(self) -> None:
        """Starts the processing of authorize URLs.

        Will open a webbrowser for the first one; others are then handled as redirects
        from each callback URL.
        """
        if not self._initialised:
            logger.critical(
                "OIDC callback endpoint has not been initialised. "
                "Unable to complete OIDC authentication."
            )
            raise RuntimeError(
                "OIDC callback endpoint has not been initialised. "
                "Unable to complete OIDC authentication."
            )

        first_url = next(self._urls_iter)

        # Open first authorization URL (it will autoredirect to the rest)
        logger.info(
            f"Attempting to open browser. "
            f"Running a headless client? "
            f"You'll need to open this link in a browser: {first_url}"
        )
        webbrowser.open(first_url)

    async def process_callback(self, request: AioRequest) -> AioResponse:
        """Process each authorize callback, extracting the auth code.

        Will automatically redirect to the next authorize URL in the chain until
        all have been processed.
        """
        if not self._initialised:
            logger.critical(
                "OIDC callback endpoint has not been initialised. "
                "Unable to complete OIDC authentication."
            )
            raise RuntimeError(
                "OIDC callback endpoint has not been initialised. "
                "Unable to complete OIDC authentication."
            )

        # Extract and store response details from the callback URL. The `state`
        # parameter is used to identify which pod this auth code is intended for,
        # as well as to check that the response is valid.
        query_dict = request.query
        auth_code = query_dict["code"]
        state = query_dict["state"]
        pod_id = self._state_to_id[state]
        self._responses[pod_id] = _OIDCAuthEndpointResponse(auth_code, state)

        # Get next /authorize URL to go to
        try:
            next_url = next(self._urls_iter)
            # aiohttp uses an exception to return a redirect response
            raise web.HTTPFound(next_url)
        except StopIteration:
            # All URLs visited, mark done
            self.all_done.set()
            return AioResponse(
                text="You've now proven your identity to all pods involved in the "
                "task. You can close this tab."
            )

    async def get_responses(self) -> Dict[str, _OIDCAuthEndpointResponse]:
        """Get the responses to each authorize call.

        Will wait until all calls have completed.
        """
        await self.all_done.wait()
        return self._responses


def _get_urlsafe_hash(s: str, initial_encoding: str = "utf-8") -> str:
    """Generate a URL-safe hash of a string.

    Encodes the string initially using the provided encoding, finds the sha256
    hash of this, b64-urlencodes the hash digest, and removes the padding "=" that
    python keeps in there.
    """
    encoded_s: bytes = s.encode(initial_encoding)
    sha256_hash: bytes = hashlib.sha256(encoded_s).digest()
    b64_hash: bytes = base64.urlsafe_b64encode(sha256_hash)
    b64_hash_str: str = b64_hash.decode("utf-8").replace("=", "")
    return b64_hash_str


def _verify_client_ids(
    oidc_details: Dict[str, _OIDCClientID], expected_auth_domain: str
) -> None:
    """Checks that all received client IDs match the expected one.

    Raises:
        AuthEnvironmentError:
            If the Client IDs for this auth environment do not match those
            received from the pods.
    """
    our_auth_env = _get_auth_environment()

    if expected_auth_domain != our_auth_env.auth_domain:
        raise ValueError(
            f"Mismatch between setup auth environment and current: "
            f"expected {expected_auth_domain}, got {our_auth_env.auth_domain}"
        )

    errors = [
        (pod_id, details)
        for pod_id, details in oidc_details.items()
        if details.client_id != our_auth_env.client_id
    ]
    if errors:
        errors_str = "; ".join(
            [
                f"Pod ID = {pod_id}, Client ID = {details.client_id}"
                for pod_id, details in errors
            ]
        )
        raise AuthEnvironmentError(
            f"Authorisation environments do not match. "
            f'Expected client_id "{our_auth_env.client_id}" but the following '
            f"pods mismatched: {errors_str}"
        )


class _OIDCAuthFlowChallengeHandler(_ResponseHandler, _HasWebServer):
    """Perform OIDC flow from the modeller-side.

    This flow is based on:
    https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow-with-proof-key-for-code-exchange-pkce  # noqa: B950
    """

    def __init__(
        self,
        auth_domain: str = _PRODUCTION_AUTH_DOMAIN,
        scopes: str = _OIDC_IDENTITY_VERIFICATION_SCOPES,
        audience: str = _OIDC_IDENTITY_VERIFICATION_AUDIENCE,
    ):
        self._auth_domain = auth_domain
        self._authorize_endpoint = f"https://{self._auth_domain}/authorize"
        self._redirect_uri = (
            f"http://localhost:{_BITFOUNT_MODELLER_PORT}{_CALLBACK_ROUTE}"
        )

        self._scopes: str = scopes
        self._audience: str = audience

        # Create web application for callbacks
        self._oidc_endpoint = _OIDCWebEndpoint()
        app = Application()
        app.add_routes([web.get(_CALLBACK_ROUTE, self._oidc_endpoint.process_callback)])
        self._runner = AppRunner(app)

        # This task is used to ensure we don't try to
        # perform authentication before the server has started
        self._server_start_task: Task

    def start_server(self) -> Task:
        """Sets up and starts the web server."""
        self._server_start_task = create_task(self._start())
        return self._server_start_task

    async def _start(self) -> None:
        """Sets up and starts the web server."""
        logger.debug("Starting OIDC web server")
        host = "localhost"
        port = _BITFOUNT_MODELLER_PORT

        # Try to create web server for endpoint, failing out if binding to the
        # host and port isn't possible.
        try:
            await self._runner.setup()
            # This site is accessed by the modeller (who is running this code)
            # in their own browser.
            # Their OIDC IdP provider will redirect them to:
            # `http://localhost:{BITFOUNT_MODELLER_PORT}/{_CALLBACK_ROUTE}`
            # So they should not have any issues accessing it.
            site = TCPSite(self._runner, host, port)
            # wait_for() use ensures we don't hang indefinitely on waiting for the
            # port to be available
            await asyncio.wait_for(site.start(), _PORT_WAIT_TIMEOUT)
        except TimeoutError as e:
            logger.critical(
                f"Timeout reached whilst trying to bind OIDC web endpoint to "
                f"http://{host}:{port}"
            )
            raise e
        except OSError as e:
            # Raises OSError: [Errno 48] if address already in use
            logger.critical(f"Unable to bind OIDC web endpoint to http://{host}:{port}")
            raise e

        logger.debug("OIDC web server started successfully")

    async def handle(self, modeller_mailbox: _ModellerMailbox) -> None:
        """Receive and perform OIDC verification."""
        try:
            # Check that start_server() has been called
            try:
                self._server_start_task
            except AttributeError:
                raise RuntimeError(
                    "OIDC server has not been started; "
                    "ensure start_server() has been called."
                )

            # Await Client ID information from pods
            oidc_details = await modeller_mailbox.get_oidc_client_ids()

            # Verify Client IDs are as expected
            self._verify_client_ids(oidc_details)

            # Generate /authorize URL for each authorization request
            to_send_dicts: DefaultDict[str, _JSONDict] = defaultdict(dict)
            urls: Dict[str, str] = {}
            states: Dict[str, str] = {}
            for pod_id, details in oidc_details.items():
                # Generate code verifier
                # Should be between 43 and 128 characters; recommended to use
                # base64url-encoding which secrets.token_urlsafe does for us. 60 bytes
                # will produce ~78 characters.
                # See: https://datatracker.ietf.org/doc/html/rfc7636#section-4.1
                code_verifier: str = secrets.token_urlsafe(nbytes=60)
                # Add to details
                to_send_dicts[pod_id]["code_verifier"] = code_verifier

                # Generate code challenge
                # This is the b64url-encoded SHA256 hash of the ascii-encoded
                # code_verifier.
                # We also encode the bytes-output of the b64 encoding with UTF-8 to
                # make it a string.
                # See: https://datatracker.ietf.org/doc/html/rfc7636#section-4.2
                code_challenge: str = _get_urlsafe_hash(
                    code_verifier, initial_encoding="ascii"
                )

                # Request PKCE authorization
                # Generate random state
                state = secrets.token_hex(nbytes=32)
                params = urlencode(
                    {
                        "audience": self._audience,
                        "scope": self._scopes,
                        "response_type": "code",
                        "client_id": details.client_id,
                        "state": state,
                        "redirect_uri": self._redirect_uri,
                        "code_challenge_method": "S256",
                        "code_challenge": code_challenge,
                    }
                )
                url = self._authorize_endpoint + f"?{params}"

                urls[pod_id] = url
                states[pod_id] = state

            # Set generated URLs on web endpoint for later iteration
            urls_list: List[str] = list(urls.values())
            self._oidc_endpoint.initialise(urls_list, states)

            # Wait for server if necessary
            if not self._server_start_task.done():
                logger.info("Waiting for OIDC challenge handler to start")
                await self._server_start_task
            # If an exception was thrown in the task,
            # Task.result() will re-raise it for us.
            self._server_start_task.result()

            # Start URL processing
            self._oidc_endpoint.start_processing()

            # Wait for all authorization to be done
            responses = await self._oidc_endpoint.get_responses()

            # Validate the states are the same
            for pod_id, response in responses.items():
                if response.state != states[pod_id]:
                    raise ValueError(
                        f"Unable to validate response intended for {pod_id}"
                    )

            # Add auth_code and redirect_uri to details
            for pod_id in oidc_details:
                to_send_dicts[pod_id]["auth_code"] = responses[pod_id].auth_code
                to_send_dicts[pod_id]["redirect_uri"] = self._redirect_uri

            # Send details to pods
            to_send: Dict[str, _OIDCAuthFlowResponse] = {
                pod_id: _OIDCAuthFlowResponse(**d)
                for pod_id, d in to_send_dicts.items()
            }
            await modeller_mailbox.send_oidc_auth_flow_responses(to_send)
        finally:
            await self.stop_server()

    def _verify_client_ids(self, oidc_details: Dict[str, _OIDCClientID]) -> None:
        """Checks that all received client IDs match the expected one.

        Raises:
            AuthEnvironmentError:
                If the Client IDs for this auth environment do not match those
                received from the pods.
        """
        return _verify_client_ids(oidc_details, expected_auth_domain=self._auth_domain)

    async def stop_server(self) -> None:
        """Stop the web server."""
        await self._runner.cleanup()


class _OIDCDeviceCodeHandler(_ResponseHandler):
    """OIDC Device Authorisation Flow from modeller-side.

    See: https://auth0.com/docs/get-started/authentication-and-authorization-flow/device-authorization-flow  # noqa: B950
    """

    def __init__(
        self,
        auth_domain: str = _PRODUCTION_AUTH_DOMAIN,
        scopes: str = _OIDC_IDENTITY_VERIFICATION_SCOPES,
        audience: str = _OIDC_IDENTITY_VERIFICATION_AUDIENCE,
    ):
        self._auth_domain = auth_domain
        self._device_code_endpoint = f"https://{self._auth_domain}/oauth/device/code"

        self._scopes: str = scopes
        self._audience: str = audience

    async def handle(self, modeller_mailbox: _ModellerMailbox) -> None:
        """See parent class for more information."""
        logger.debug("Getting oidc client ids")
        # Await Client ID information from pods
        oidc_details = await modeller_mailbox.get_oidc_client_ids()

        # Verify Client IDs are as expected
        self._verify_client_ids(oidc_details)

        # Get device code details for each pod
        logger.info("Retrieving device codes for identity verification.")
        device_codes: Dict[str, _DeviceCodeDetailsPair] = {}
        for pod_id, details in oidc_details.items():
            device_codes[pod_id] = self._get_device_code(details.client_id)

        # Send relevant details to pods
        logger.info("Sending identity verification details to pods")
        await modeller_mailbox.send_oidc_device_code_responses(
            {pod_id: details.pod_details for pod_id, details in device_codes.items()}
        )

        # Get user to approve the access requests
        # Print details about what's happening
        if len(device_codes) > 1:
            browser_opened_str = (
                "Browser windows will be opened, please login and confirm "
                "identity verification access for these pods."
            )
            cli_str = (
                "If no or not all browser windows open, then please visit "
                "the following URLs:"
            )
        else:
            browser_opened_str = (
                "A browser window will be opened, please login and confirm "
                "identity verification access for the pod."
            )
            cli_str = (
                "If a browser window is not opened, then please visit the "
                "following URL:"
            )
        print(browser_opened_str)
        print(cli_str)

        # Print user code details so they know what to verify and the URIs so they
        # can manually open them if needed.
        to_open = []
        for pod_id, device_code_details in device_codes.items():
            verification_uri_complete = (
                device_code_details.modeller_details.verification_uri_complete
            )
            user_code = device_code_details.modeller_details.user_code

            print(f"\tFor {pod_id}: code: {user_code}; {verification_uri_complete}")
            to_open.append(verification_uri_complete)

        # Give user time to see the codes, so they know what to verify
        await asyncio.sleep(1)

        # Open all URIs
        for uri in to_open:
            webbrowser.open_new_tab(uri)

    def _verify_client_ids(self, oidc_details: Dict[str, _OIDCClientID]) -> None:
        """Checks that all received client IDs match the expected one.

        Raises:
            AuthEnvironmentError:
                If the Client IDs for this auth environment do not match those
                received from the pods.
        """
        return _verify_client_ids(oidc_details, expected_auth_domain=self._auth_domain)

    def _get_device_code(self, client_id: str) -> _DeviceCodeDetailsPair:
        logger.debug(f"Retrieving device code for client id: {client_id}")
        # Make request and raise for error if needed
        request_data: _DeviceCodeRequestDict = {
            "audience": self._audience,
            "scope": self._scopes,
            "client_id": client_id,
        }
        response = web_utils.post(
            self._device_code_endpoint,
            data=request_data,
        )

        # Handle 4XX/5XX errors
        response.raise_for_status()

        # Handle other non-200 errors
        if (status_code := response.status_code) != 200:
            raise HTTPError(
                f"Unexpected response: status_code = {status_code}; {response.text}"
            )

        response_json: _DeviceCodeResponseJSON = response.json()

        # Calculate expiration time
        expires_at = datetime.now(timezone.utc) + timedelta(
            seconds=response_json["expires_in"]
        )

        return _DeviceCodeDetailsPair(
            _PodDeviceCodeDetails(
                device_code=response_json["device_code"],
                expires_at=expires_at,
                interval=response_json["interval"],
            ),
            _ModellerDeviceCodeDetails(
                user_code=response_json["user_code"],
                verification_uri=response_json["verification_uri"],
                verification_uri_complete=response_json["verification_uri_complete"],
            ),
        )



"""Type hints, enums and protocols for the Bitfount libraries."""
from __future__ import annotations

import os
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    MutableMapping,
    NewType,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    runtime_checkable,
)

import marshmallow
from marshmallow import fields
import numpy as np
from pandas._typing import Dtype

from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.data.dataloaders import BitfountDataLoader
    from bitfount.data.datasources.base_source import BaseSource
    from bitfount.data.datastructure import DataStructure
    from bitfount.data.schema import BitfountSchema
    from bitfount.federated.model_reference import BitfountModelReference
    from bitfount.metrics import Metric
    from bitfount.models.base_models import ModelContext

__all__: List[str] = ["DistributedModelProtocol"]

# Tensor dtype type variable
T_DTYPE = TypeVar("T_DTYPE", covariant=True)


# TensorLike protocol and TensorLike composite types
class _TensorLike(Protocol):
    """Protocol defining what methods and operations a Generic Tensor can perform."""

    def __mul__(self: _TensorLike, other: Any) -> _TensorLike:
        """Multiplies the self tensor with the other tensor and returns the result."""
        ...

    def __sub__(self: _TensorLike, other: Any) -> _TensorLike:
        """Subtracts the other tensor from the self tensor and returns the result."""
        ...

    def squeeze(self: _TensorLike, axis: Optional[Any] = None) -> _TensorLike:
        """Returns a tensor with all the dimensions of input of size 1 removed."""
        ...


# Weight update types
_SerializedWeights = Mapping[str, List[float]]
_Residuals = Mapping[str, _TensorLike]
_Weights = Mapping[str, _TensorLike]


# Schema dtypes
_DtypesValues = Union[Dtype, np.dtype]
_Dtypes = Dict[str, _DtypesValues]


@delegates()
# DistributedModel protocol and types
@runtime_checkable
class DistributedModelProtocol(Protocol):
    """Federated Model structural type.

    This protocol should be implemented by classes that inherit from either
    `BitfountModel`, or both of `_BaseModel` and `DistributedModelMixIn`.
    """

    class_name: str
    datastructure: DataStructure
    schema: BitfountSchema  # TODO: [NO_TICKET: To discuss about the schema being here.] # noqa: B950
    # Type hints below indicate that one of either `epochs` or `steps` needs to be
    # supplied by the mixed-in class or other classes in the inheritance hierarchy
    epochs: Optional[int]
    steps: Optional[int]
    _total_num_batches_trained: int
    # Denotes the Pod the model is running in (if any)
    pod_identifier: Optional[str]
    # Denotes if param_clipping params are given to the model.
    param_clipping: Optional[Dict[str, int]]
    metrics: Optional[MutableMapping[str, Metric]]
    fields_dict: ClassVar[T_FIELDS_DICT]
    nested_fields: ClassVar[T_NESTED_FIELDS]

    def tensor_precision(self) -> T_DTYPE:
        """Defined in DistributedModelMixIn."""

    def get_param_states(self) -> _Weights:
        """Defined in DistributedModelMixIn."""

    def apply_weight_updates(self, weight_updates: Sequence[_Weights]) -> _Weights:
        """Defined in DistributedModelMixIn."""

    def update_params(self, new_model_params: _Weights) -> None:
        """Defined in DistributedModelMixIn."""

    def serialize_params(self, weights: _Weights) -> _SerializedWeights:
        """Defined in DistributedModelMixIn."""

    def deserialize_params(self, serialized_weights: _SerializedWeights) -> _Weights:
        """Defined in DistributedModelMixIn."""

    def diff_params(self, old_params: _Weights, new_params: _Weights) -> _Residuals:
        """Defined in DistributedModelMixIn."""

    def set_model_training_iterations(self, iterations: int) -> None:
        """Defined in DistributedModelMixIn."""

    def reset_trainer(self) -> None:
        """Defined in DistributedModelMixIn."""

    def set_pod_identifier(self, pod_identifier: str) -> None:
        """Defined in DistributedModelMixIn."""

    def fit(
        self,
        data: Optional[BaseSource] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        pod_identifiers: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Optional[Dict[str, str]]:
        """Defined in DistributedModelMixIn."""

    def log_(self, name: str, value: Any, **kwargs: Any) -> Any:
        """Defined in DistributedModelMixIn."""

    @property
    def initialised(self) -> bool:
        """Defined in _BaseModel."""

    def initialise_model(
        self,
        data: Optional[BaseSource] = None,
        context: Optional[ModelContext] = None,
    ) -> None:
        """Defined in _BaseModel."""

    def evaluate(
        self,
        test_dl: Optional[BitfountDataLoader] = None,
        pod_identifiers: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Dict[str, float]]:
        """Defined in _BaseModel."""

    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Defined in _BaseModel."""

    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Defined in _BaseModel."""


T_FIELDS_DICT = Dict[str, marshmallow.fields.Field]
T_NESTED_FIELDS = Dict[str, Mapping[str, Any]]
T_OPTIONAL_NESTED = Optional[List[str]]


class _BaseSerializableObjectMixIn:
    """The base class from which all serializable objects should inherit from.

    Attributes:
        fields_dict: A dictionary mapping all attributes that will be serialized
            in the class to their marshamllow field type. (e.g.
            fields_dict = {"class_name": fields.Str()}).
        nested_fields: A dictionary mapping all nested attributes to a registry
            that contains class names mapped to the respective classes.
            (e.g. nested_fields = {"datastructure": datastructure.registry})
        optional_nested: A list of strings that contains the keys from the
            nested classes that are optional for the class that needs to
            be serialized. (e.g. optional_nested = ["datastructure"]
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {"class_name": fields.Str()}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}
    optional_nested: ClassVar[T_OPTIONAL_NESTED] = None


if TYPE_CHECKING:
    _DistributedModelTypeOrReference = Union[
        DistributedModelProtocol, BitfountModelReference
    ]

# Serialization Types
_JSONDict = Dict[str, Any]  # A JSON-esque dictionary that is serializable

# Common Types
_StrAnyDict = Dict[str, Any]  # A dictionary with string keys and any value types

# S3 Interaction Types
_S3PresignedPOSTURL = NewType("_S3PresignedPOSTURL", str)
# HTTPX explicitly expects a `dict` rather than a `mapping`
_S3PresignedPOSTFields = NewType("_S3PresignedPOSTFields", Dict[str, str])
_S3PresignedURL = NewType("_S3PresignedURL", str)  # for GET requests

# SAML Types
_SAMLResponse = Mapping[str, Any]



"""Pytest config and hooks for tests."""
# Import fixtures
import logging
import multiprocessing
import platform
import sys
from typing import Any, Generator, List, Set
from unittest.mock import Mock

# noinspection PyUnresolvedReferences
from _pytest.monkeypatch import MonkeyPatch
import pandas as pd
import psycopg
import pytest

# noinspection PyUnresolvedReferences
from pytest import fixture
from pytest_mock import MockerFixture
import sqlalchemy

from tests.utils.fixtures import *  # noqa: F401, F403
from tests.utils.helper import (
    backend_test,
    create_dataset,
    create_dataset_pii,
    end_to_end_mocks_test,
    end_to_end_test,
    integration_test,
    tutorial_test,
    unit_test,
)

logger = logging.getLogger(__name__)

_REQUIRED_MARKERS: Set[str] = {
    tutorial_test.name,
    end_to_end_mocks_test.name,
    end_to_end_test.name,
    integration_test.name,
    unit_test.name,
}

_BACKEND_PACKAGE_NAME: str = "backends"
_BACKEND_MARKER: str = backend_test.name

_BACKEND_MARK_EXCLUSIONS: Set[str] = set(
    [
        # Add node IDs that should be excluded from requiring the backend_test marker
    ]
)


@fixture(autouse=True, scope="session")
def multiprocessing_start_method() -> None:
    """Sets multiprocessing start method to 'fork' rather than 'spawn'.

    This can only be set once and it has to be before the library is used anywhere which
    is why it must be an autouse session scope fixture.

    Required for Python 3.8 on Macs when using Flask but ensures consistency between
    Unix platforms ("fork" is the default on non-macOS Unix).

    See: https://github.com/pytest-dev/pytest-flask/issues/104
    """
    if platform.system() == "Windows":
        multiprocessing.set_start_method("spawn")
    else:
        multiprocessing.set_start_method("fork")


@fixture(autouse=True, scope="session")
def mac_m1_env_fix(monkeypatch_session_scope: MonkeyPatch) -> None:
    """This empirically fixes errors on this test for Macs with M1 chip.

    Setting the `no_proxy` environment variable disables network proxy lookups. The
    solution was taken from a much older python bug which also fixes this for some
    reason: https://bugs.python.org/issue28342
    """
    if sys.platform == "darwin" and "ARM64" in platform.version():
        logger.info("M1 Mac detected, setting envvar 'no_proxy' to '*'")
        # Checking for 'ARM64'in `platform.version()` appears to be the only way
        # to ensure we find Macs with the M1 chip even when the python interpreter
        # is using Rosetta
        monkeypatch_session_scope.setenv("no_proxy", "*")


@fixture
def mock_pandas_read_sql_query(monkeypatch: MonkeyPatch) -> None:
    """Pandas `read_sql_query()` mocked."""
    data = create_dataset()

    def get_df(**_kwargs: Any) -> pd.DataFrame:
        return data

    monkeypatch.setattr(pd, "read_sql_query", get_df)


@fixture
def mock_inspector(mocker: MockerFixture) -> Generator[Mock, None, None]:
    """Automatically mocks sqlalchemy inspector and yields mocked object."""
    mock_inspector = Mock(
        default_schema_name="public", spec=sqlalchemy.engine.Inspector
    )
    mock_inspector.get_schema_names.return_value = ["public"]
    mock_inspector.get_table_names.return_value = ["dummy_data", "dummy_data_2"]
    mocker.patch("bitfount.data.utils.inspect", return_value=mock_inspector)
    yield mock_inspector


@fixture
def mock_engine(mock_inspector: Mock) -> Generator[Mock, None, None]:
    """Returns mock sqlalchemy engine."""
    yield Mock(spec=sqlalchemy.engine.base.Engine)


@fixture
def db_session(
    postgresql: psycopg.Connection,
) -> Generator[sqlalchemy.engine.base.Engine, None, None]:
    """Creates a dummy postgres database connection.

    This fixture should only be used for integration or end-to-end tests.
    """
    connection = (
        f"postgresql+psycopg2://{postgresql.info.user}:"
        f"@{postgresql.info.host}:{postgresql.info.port}/{postgresql.info.dbname}"
    )
    engine = sqlalchemy.create_engine(connection)

    df = create_dataset()
    df2 = create_dataset()
    # The tables should never already exist in the database so we set it to fail
    # if it does to catch any potential setup errors.
    df.to_sql("dummy_data", engine, if_exists="fail", index=False)
    df2.to_sql("dummy_data_2", engine, if_exists="fail", index=False)

    yield engine


@fixture
def db_session_pii(
    postgresql: psycopg.Connection,
) -> Generator[sqlalchemy.engine.base.Engine, None, None]:
    """Creates a dummy postgres database connection."""
    connection = (
        f"postgresql+psycopg2://{postgresql.info.user}:"
        f"@{postgresql.info.host}:{postgresql.info.port}/{postgresql.info.dbname}"
    )
    engine = sqlalchemy.create_engine(connection)

    # Create first table, drop height
    df = create_dataset_pii()
    df.drop("height", inplace=True, axis=1)
    # Create second, different table including height
    df2 = create_dataset_pii()
    df2.drop("age", inplace=True, axis=1)
    df2.drop("weight", inplace=True, axis=1)
    df2.drop("exercise", inplace=True, axis=1)

    # The tables should never already exist in the database so we set it to fail
    # if it does to catch any potential setup errors.
    df.to_sql("dummy_data", index=False, con=engine, if_exists="fail")
    df2.to_sql("dummy_data_2", index=False, con=engine, if_exists="fail")

    yield engine


@pytest.fixture(scope="module")
def monkeypatch_module_scope() -> Generator[MonkeyPatch, None, None]:
    """Module-scoped monkeypatch."""
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="session")
def monkeypatch_session_scope() -> Generator[MonkeyPatch, None, None]:
    """Session-scoped monkeypatch."""
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


# Create hook to ensure that all tests are correctly marked
def pytest_collection_modifyitems(items: List[pytest.Item]) -> None:
    """Hooks into the pytest test collection to ensure marker criteria met."""
    no_marks: List[str] = []
    missing_backend_marks: List[str] = []

    for item in items:
        item_id = item.nodeid
        item_markers = set(m.name for m in item.iter_markers())

        # Check each found test has a required mark
        if _REQUIRED_MARKERS.isdisjoint(item_markers):
            no_marks.append(item_id)

        # Check backend tests have backend mark
        if _BACKEND_PACKAGE_NAME in item_id:
            if (
                _BACKEND_MARKER not in item_markers
                and item_id not in _BACKEND_MARK_EXCLUSIONS
            ):
                missing_backend_marks.append(item_id)

        # Add marker to skip postgresql tests on Windows
        # TODO: [BIT-2011] Remove this once we can support these tests on windows
        if "postgresql" in getattr(item, "fixturenames", ()):
            item.add_marker(
                pytest.mark.skipif(
                    condition=platform.system() == "Windows",
                    reason="'pytest_postgresql' fixture not supported on Windows.",
                )
            )

    # Error out if any tests were found without marks
    if no_marks:
        no_marks_str = "    " + "\n    ".join(no_marks)
        raise ValueError(
            f"All tests require one of tutorial_test, end_to_end_test, "
            f"end_to_end_mocks_test, integration_test or unit_test as markers; "
            f"some tests have none:\n{no_marks_str}"
        )

    # Error out if any backend tests missing mark
    if missing_backend_marks:
        missing_backend_marks_str = "    " + "\n    ".join(missing_backend_marks)
        raise ValueError(
            f"The following tests may require the backend_test marker; "
            f"add this or mark them as excluded in tests/conftest.py:"
            f"\n{missing_backend_marks_str}"
        )



"""Fixtures and other pytest specific components for the transformation tests."""
from _pytest.monkeypatch import MonkeyPatch
from pytest import fixture


@fixture
def fake_uuid(monkeypatch: MonkeyPatch) -> None:
    """Patch the uuid4().hex calls in base_transformation."""
    import bitfount.transformations.base_transformation as base_transformation

    class FakeUUID:
        hex = "FAKE_HEX"

    def fake_uuid() -> FakeUUID:
        return FakeUUID()

    monkeypatch.setattr(base_transformation, "uuid4", fake_uuid)


@fixture
def empty_registry(monkeypatch: MonkeyPatch) -> None:
    """Ensure an empty registry in base_transformation."""
    import bitfount.transformations.base_transformation as base_transformation

    monkeypatch.setattr(base_transformation, "TRANSFORMATION_REGISTRY", {})



"""Testcases for all classes in bitfount/utils.py."""
import inspect
import logging
from pathlib import Path
import re
import sys
from typing import Any, Generator, List, Literal, NoReturn, Tuple, Union, cast

from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
import docstring_parser.common
import numpy as np
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
from testbook import testbook

from bitfount.utils import (
    _add_this_to_list,
    _find_logger_filenames,
    _full_traceback,
    _get_mb_from_bytes,
    _get_object_source_code,
    _handle_fatal_error,
    _inspect_get_file,
    _is_notebook,
    _powinv,
    _powmod,
    delegates,
    one_hot_encode_list,
)
from tests.utils import PytestRequest
from tests.utils.helper import get_critical_logs, get_debug_logs, unit_test


@unit_test
class TestIsNotebook:
    """Tests is_notebook()."""

    def test_notebook(self) -> None:
        """Tests fit_all_datasets()."""
        return_val = _is_notebook()
        assert not return_val


@unit_test
class TestAddThisToList:
    """Tests add_this_to_list()."""

    def test_add_duplicate(self) -> None:
        """Tests adding duplicate."""
        this = 1
        lst = [1, 2, 3]
        lst = _add_this_to_list(this, lst)
        assert lst == [1, 2, 3]

    def test_add_none(self) -> None:
        """Tests adding none."""
        this = None
        lst = [1, 2, 3]
        lst = _add_this_to_list(this, lst)
        assert lst == [1, 2, 3]

    def test_add_new_value(self) -> None:
        """Tests adding new value."""
        this = 4
        lst = [1, 2, 3]
        lst = _add_this_to_list(this, lst)
        assert lst == [1, 2, 3, 4]

    def test_add_list(self) -> None:
        """Tests adding list."""
        this = [4]
        lst = [1, 2, 3]
        lst = _add_this_to_list(this, lst)
        assert lst == [1, 2, 3, 4]


@unit_test
class TestOneHotEncodeList:
    """Tests one_hot_encode_list."""

    @staticmethod
    def data(dims: int) -> Union[List[int], List[List[int]]]:
        """Fixture of input list (or 2D list) of integers."""
        if dims == 1:
            return [0, 1, 2, 1]
        elif dims == 2:
            return [[0, 1], [1, 2], [2, 1], [1, 0]]
        else:
            raise ValueError(f"Unsupported dimension: {dims}")

    @staticmethod
    def expected(dims: int) -> np.ndarray:
        """Fixture of expected OHE output array."""
        if dims == 1:
            return np.array(
                [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 0]], dtype=np.uint8
            )
        elif dims == 2:
            return np.array(
                [
                    [1, 0, 0, 0, 1, 0],
                    [0, 1, 0, 0, 0, 1],
                    [0, 0, 1, 0, 1, 0],
                    [0, 1, 0, 1, 0, 0],
                ],
                dtype=np.uint8,
            )
        else:
            raise ValueError(f"Unsupported dimension: {dims}")

    @fixture(params=[1, 2], ids=["1D", "2D"])
    def data_and_expected(
        self, request: PytestRequest
    ) -> Tuple[Union[List[int], List[List[int]]], np.ndarray]:
        """Fixture combining data and expected for different dimensions."""
        return self.data(request.param), self.expected(request.param)

    def test_one_hot_encode_int_list(
        self, data_and_expected: Tuple[Union[List[int], List[List[int]]], np.ndarray]
    ) -> None:
        """Tests one_hot_encode_list for int list."""
        data, expected = data_and_expected
        ohe = one_hot_encode_list(data)
        assert np.array_equal(ohe, expected)

    def test_one_hot_encode_array_list(
        self, data_and_expected: Tuple[Union[List[int], List[List[int]]], np.ndarray]
    ) -> None:
        """Tests one_hot_encode_list for array list."""
        data, expected = data_and_expected
        data_arrays = [np.array(i) for i in data]
        assert isinstance(data_arrays, list)
        assert isinstance(data_arrays[0], np.ndarray)
        ohe = one_hot_encode_list(data_arrays)
        assert np.array_equal(ohe, expected)

    def test_one_hot_encode_array(
        self, data_and_expected: Tuple[Union[List[int], List[List[int]]], np.ndarray]
    ) -> None:
        """Tests one_hot_encode_list for array."""
        data, expected = data_and_expected
        data_array = np.asarray(data)
        assert isinstance(data_array, np.ndarray)
        ohe = one_hot_encode_list(data_array)
        assert np.array_equal(ohe, expected)

    def test_one_hot_encode_fails_3D(self) -> None:
        """Tests one hot encoding fails for 3D data."""
        data = [[[1], [2], [3]]]
        with pytest.raises(
            ValueError,
            match="Incorrect number of dimensions for one-hot encoding; "
            "expected 1 or 2, got 3",
        ):
            one_hot_encode_list(data)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950

    def test_one_hot_encode_fails_0D(self) -> None:
        """Tests one hot encoding fails for scalar data."""
        data = 1
        with pytest.raises(
            ValueError,
            match="Incorrect number of dimensions for one-hot encoding; "
            "expected 1 or 2, got 0",
        ):
            one_hot_encode_list(data)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950


@unit_test
def test_get_mb_from_bytes() -> None:
    """Tests get_mb_from_bytes works correctly."""
    # Test with whole number of MB bytes
    whole_mb = 2 * 1024 * 1024  # 2MB
    mb = _get_mb_from_bytes(whole_mb)
    assert mb.whole == 2
    assert mb.fractional == 2.0

    # Test with non-whole number of MB bytes
    non_whole_mb = whole_mb + 1
    mb = _get_mb_from_bytes(non_whole_mb)
    assert mb.whole == 2
    assert mb.fractional == non_whole_mb / (1024 * 1024)


@unit_test
class TestFatalErrorHandling:
    """Tests for fatal error handling functions."""

    def test__find_logger_filenames(self, tmp_path: Path) -> None:
        """Tests that _find_logger_filenames works correctly.

        Tests that it finds multiple filenames, but only those at the correct level.
        """
        # Set a base logger with INFO and a file handler for CRITICAL
        logger_1 = logging.getLogger("logger_1")
        logger_1.setLevel(logging.INFO)
        path_1 = tmp_path / "log_1.log"
        handler_1 = logging.FileHandler(path_1)
        handler_1.setLevel(logging.CRITICAL)
        logger_1.addHandler(handler_1)
        # Need to treat _this_ as the root logger as pytest messes with the root
        # logger handlers
        logger_1.parent = None

        # Set a child logger which will inherit the level from the base logger
        # and has a separate file handler at WARNING
        logger_2 = logging.getLogger("logger_1.logger_2")
        path_2 = tmp_path / "log_2.log"
        handler_2 = logging.FileHandler(path_2)
        handler_2.setLevel(logging.WARNING)
        logger_2.addHandler(handler_2)

        critical_log_files = _find_logger_filenames(
            logger_2, cast(Literal[50], logging.CRITICAL)
        )
        warning_log_files = _find_logger_filenames(
            logger_2, cast(Literal[30], logging.WARNING)
        )
        info_log_files = _find_logger_filenames(
            logger_2, cast(Literal[20], logging.INFO)
        )
        debug_filenames = _find_logger_filenames(
            logger_2, cast(Literal[10], logging.DEBUG)
        )

        # CRITICAL should contain both files, in reverse order
        assert critical_log_files == [str(path_2), str(path_1)]
        # WARNING should only contain the second file
        assert warning_log_files == [str(path_2)]
        # INFO should contain no files (i.e. be None) as no file handlers exist
        # for that level
        assert info_log_files is None
        # DEBUG should contain no files (i.e. be None) as the logger is not enabled
        # for that level
        assert debug_filenames is None

    @pytest.mark.parametrize(argnames="traceback_prelimited", argvalues=(True, False))
    def test__full_traceback_context_manager(
        self,
        monkeypatch: MonkeyPatch,
        traceback_prelimited: bool,
    ) -> None:
        """Tests that _full_traceback correctly manipulates sys.tracebacklimit."""
        limit = 10

        # Temporarily delete any previously set tracebacklimit
        monkeypatch.delattr("sys.tracebacklimit", raising=False)

        # Check condition before entering context manager
        if traceback_prelimited:
            sys.tracebacklimit = limit
            assert sys.tracebacklimit == limit
        else:
            assert not hasattr(sys, "tracebacklimit")

        with _full_traceback():
            # Should not have tracebacklimit set
            assert not hasattr(sys, "tracebacklimit")

        # Check post-condition after context manager
        if traceback_prelimited:
            assert sys.tracebacklimit == limit
        else:
            assert not hasattr(sys, "tracebacklimit")

    @pytest.fixture
    def temp_tb_limit(self) -> Generator[int, None, None]:
        """Temporarily manipulate tracebacklimit to ensure reset."""
        orig = sys.tracebacklimit
        temp_tb_limit = orig + 1
        sys.tracebacklimit = temp_tb_limit
        try:
            yield temp_tb_limit
        finally:
            sys.tracebacklimit = orig

    @pytest.mark.parametrize(
        argnames="cli_mode",
        argvalues=(True, False),
        ids=lambda x: f"cli_mode={x}",
    )
    @pytest.mark.parametrize(
        argnames="dev_mode",
        argvalues=(True, False),
        ids=lambda x: f"dev_mode={x}",
    )
    def test__handle_fatal_error(
        self,
        caplog: LogCaptureFixture,
        cli_mode: bool,
        dev_mode: bool,
        mocker: MockerFixture,
        monkeypatch: MonkeyPatch,
        temp_tb_limit: int,
    ) -> None:
        """Test _handle_fatal_error function under various configurations."""
        # Capture all logs
        caplog.set_level(logging.DEBUG)

        # Set config variables
        monkeypatch.setattr("bitfount.config._BITFOUNT_CLI_MODE", cli_mode)
        monkeypatch.setattr("bitfount.config.BITFOUNT_DEV_MODE", dev_mode)

        # Patch out log file finding
        mocker.patch(
            "bitfount.utils._find_logger_filenames",
            return_value=["/tmp/hello", "/tmp/world"],
        )

        # Create nested exception, with traceback, for us to use
        def _inner() -> NoReturn:
            raise Exception("This is the fatal error")

        def _outer() -> NoReturn:
            _inner()

        try:
            _outer()
        except Exception as exc:
            e = exc

        expected_exc_cls = SystemExit if cli_mode else Exception
        with pytest.raises(expected_exc_cls, match="^This is the fatal error$"):
            assert sys.tracebacklimit == temp_tb_limit
            _handle_fatal_error(e)

        # Check has reverted
        assert sys.tracebacklimit == temp_tb_limit

        if not dev_mode:
            # Summary in critical, full in debug
            critical_logs = get_critical_logs(caplog)
            assert (
                "This is the fatal error."
                " Full details logged to /tmp/hello, /tmp/world." == critical_logs
            )

            # Full stack trace in debug logs
            full_stack_logs = get_debug_logs(caplog, full_details=True)

            # Should be two log calls
            assert len(caplog.records) == 2
        else:
            # Full stack trace in critical logs
            full_stack_logs = get_critical_logs(caplog, full_details=True)

            # Should only be a single log call
            assert len(caplog.records) == 1

        # Full stack trace, so expect method names, keywords, line numbers, etc
        assert "This is the fatal error" in full_stack_logs
        assert "Full details logged to" not in full_stack_logs
        for func_name in ("test__handle_fatal_error", _outer.__name__, _inner.__name__):
            assert re.search(rf"line \d+, in {func_name}", full_stack_logs)
        assert 'raise Exception("This is the fatal error")' in full_stack_logs


@unit_test
class TestSourceCodeFunctionality:
    """Tests functions that are used to get the source code of an object."""

    def test__inspect_get_file_for_a_function(self, mocker: MockerFixture) -> None:
        """Tests that _inspect_get_file works correctly with a function."""
        mock_inspect_getfile = mocker.patch(
            "bitfount.utils.inspect.getfile", side_effect=inspect.getfile
        )

        def _test_func() -> None:
            """Test function."""
            pass

        file_path = _inspect_get_file(_test_func)
        assert file_path == _test_func.__code__.co_filename
        mock_inspect_getfile.assert_called_once_with(_test_func)

    def test__inspect_get_file_for_a_class(self) -> None:
        """Tests that _inspect_get_file works correctly with a class."""

        class TestClass:
            """Test class."""

            pass

        file_path = _inspect_get_file(TestClass)
        # Assert the file path is the path to this test file
        assert Path(file_path) == Path(__file__)
        assert file_path == sys.modules[TestClass.__module__].__file__

    def test__inspect_get_file_raises_a_type_error(self, mocker: MockerFixture) -> None:
        """Tests that _inspect_get_file raises a TypeError.

        This happens if the type of the object is not supported for returning the source
        code file path.
        """
        # Prevent the __module__ attribute of `TestClass` from being detected
        mocker.patch("bitfount.utils.hasattr", return_value=False)

        class TestClass:
            """Test class."""

            pass

        with pytest.raises(TypeError, match="Source for TestClass not found"):
            _inspect_get_file(TestClass)

    def test__get_object_source_code_in_a_notebook(self, mocker: MockerFixture) -> None:
        """Tests that _get_object_source_code works correctly in a notebook.

        The `_is_notebook` and `extract_symbols` functions are mocked out to simulate
        a notebook environment.
        """
        mock_is_notebook = mocker.patch(
            "bitfount.utils._is_notebook", return_value=True
        )
        mock_extract_symbols = mocker.patch(
            "bitfount.utils._get_ipython_extract_symbols",
            return_value=lambda x, y: (["test_source_code"], []),
        )

        class TestClass:
            """Test class."""

            pass

        source_code = _get_object_source_code(TestClass)
        assert source_code == "test_source_code"
        mock_is_notebook.assert_called_once()
        mock_extract_symbols.assert_called_once()

    def test_notebook_get_object_source_code(self, mocker: MockerFixture) -> None:
        """Tests that _get_object_source_code works correctly in a notebook.

        Testbook is used to execute `_get_object_source_code` inside the notebook.
        """
        with testbook(
            "tests/bitfount/resources/test_notebook.ipynb", execute=True
        ) as tb:
            # Get the source code of the dummy Blah class
            source = tb.cell_output_text(2)
            # Assert it is what is expected
            assert source == inspect.cleandoc(
                '''
                class Blah:
                    """This is just a dummy class."""

                    def blah(self) -> str:
                        """This is just a dummy method."""
                        return "dummy method return value"

                '''
            )


@unit_test
class TestGMP:
    """Tests the functions using gmpy2."""

    def test__powmod_no_gmpy(self, mocker: MockerFixture) -> None:
        """Tests that _powmod works without gmpy2."""
        mocker.patch("bitfount.utils.HAVE_GMP", False)
        py_pow = mocker.patch("bitfount.utils.pow")
        _powmod(2, 3, 11)
        py_pow.assert_called_once()

    def test__powmod_gmpy(self, mocker: MockerFixture) -> None:
        """Tests that _powmod works with gmpy2."""
        mocker.patch("bitfount.utils.HAVE_GMP", True)
        gmpy_pow = mocker.patch("bitfount.utils.gmpy2.powmod")
        _powmod(2, 3, 11)
        gmpy_pow.assert_called_once()

    def test__powinv_no_gmpy(self, mocker: MockerFixture) -> None:
        """Tests that _powinv works without gmpy2."""
        mocker.patch("bitfount.utils.HAVE_GMP", False)
        py_pow = mocker.patch("bitfount.utils.pow")
        _powinv(2, 11)
        py_pow.assert_called_once_with(2, -1, 11)

    def test__powinv_gmpy(self, mocker: MockerFixture) -> None:
        """Tests that _powinv works with gmpy2."""
        mocker.patch("bitfount.utils.HAVE_GMP", True)
        gmpy_pow = mocker.patch("bitfount.utils.gmpy2.invert")
        _powinv(2, 11)
        gmpy_pow.assert_called_once_with(2, 11)

    def test__powinv_gmpy_returns_zero(self, mocker: MockerFixture) -> None:
        """Tests that we default to pow if zero is returned."""
        mocker.patch("bitfount.utils.HAVE_GMP", True)
        mocker.patch("bitfount.utils.gmpy2.invert", return_value=0)
        py_pow = mocker.patch("bitfount.utils.pow")
        _powinv(2, 11)
        py_pow.assert_called_once_with(2, -1, 11)


@delegates(exclude_doc_meta=[docstring_parser.common.DocstringParam])
class DummyClass1:
    """Dummy class for testing.

    Args:
        arg1: An argument here.

    Attributes:
        attr1: An attribute here.

    Raises:
        error: An error here.
    """

    def __init__(self, arg1: Any, **kwargs: Any) -> None:
        pass


@delegates(keep=True, exclude_doc_meta=[docstring_parser.common.DocstringRaises])
class DummyClass2(DummyClass1):
    """Dummy class for testing.

    Args:
        arg2: An argument here.

    Attributes:
        attr2: An attribute here.

    Raises:
        error: An error here.
    """

    def __init__(self, arg2: Any, **kwargs: Any) -> None:
        pass


@unit_test
class Test_Delegates:
    """Tests the delegates decorator."""

    def test_keep_kwargs_exclude_meta(self) -> None:
        """Tests that kwargs are kept, and meta excluded."""
        dummy1 = DummyClass1(arg1="blah")
        dummy2 = DummyClass2(arg2="blah")
        # Safe to cast below as we know all classes here have docstrings.
        assert "Args" not in cast(str, dummy1.__doc__)
        assert "Attributes" not in cast(str, dummy1.__doc__)
        assert "Raises" not in cast(str, dummy2.__doc__)
        # Mypy complains that we cannot access __init__ directly,
        # however, this is only for testing purposes
        assert "arg1" in str(dummy2.__init__.__signature__)  # type: ignore[misc] # Reason: see above # noqa: B950
        assert "**kwargs" in str(dummy2.__init__.__signature__)  # type: ignore[misc] # Reason: see above # noqa: B950



"""Test SAML handling."""
import asyncio
from asyncio import Task, create_task
import re
from typing import Dict, List, Type
from unittest.mock import Mock, create_autospec

from _pytest.logging import LogCaptureFixture
from aiohttp import web
from aiohttp.pytest_plugin import AiohttpClient
from aiohttp.test_utils import TestClient
from aiohttp.web_app import Application
from aiohttp.web_runner import AppRunner, TCPSite
import pytest
from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.federated.transport.identity_verification import _BITFOUNT_MODELLER_PORT
from bitfount.federated.transport.identity_verification.saml import (
    _SAMLChallengeHandler,
    _SAMLWebEndpoint,
)
from bitfount.federated.transport.message_service import (
    _BitfountMessageType,
    _DecryptedBitfountMessage,
)
from bitfount.federated.transport.modeller_transport import (
    _ModellerMailbox,
    _WorkerMailboxDetails,
)
from bitfount.types import _SAMLResponse
from tests.utils.helper import get_critical_logs, unit_test
from tests.utils.mocks import AwaitableMock


@fixture
def saml_challenges_count() -> int:
    """Number of SAML challenges."""
    return 3


@fixture
def saml_challenges(saml_challenges_count: int) -> List[_DecryptedBitfountMessage]:
    """SAML Challenges."""
    return [
        _DecryptedBitfountMessage(
            message_type=_BitfountMessageType.SAML_REQUEST,
            body=f"saml-challenge-{i}",
            recipient="modeller",
            recipient_mailbox_id="modeller-mailbox-id",
            sender=f"sender-{i}",
            sender_mailbox_id=f"sender-mailbox-{i}",
            task_id="task-id",
        )
        for i in range(saml_challenges_count)
    ]


@fixture
def saml_responses(saml_challenges_count: int) -> List[Dict[str, str]]:
    """SAML Responses from IDP."""
    return [{"saml": f"token-{i}"} for i in range(saml_challenges_count)]


@fixture
def mailbox_details(
    saml_challenges: List[_DecryptedBitfountMessage],
) -> Dict[str, _WorkerMailboxDetails]:
    """Pod ID to Mailbox details map."""
    return {
        message.sender: _WorkerMailboxDetails(
            pod_identifier=message.sender,
            public_key=Mock(),
            mailbox_id=message.sender_mailbox_id,
            aes_encryption_key=f"aes-key-{message.sender}".encode(),
        )
        for message in saml_challenges
    }


@fixture
def idp_url() -> str:
    """IDP URL."""
    return "http://unit-test.saml-idp.bitfount.com/path?SAMLResponse="


@unit_test
class TestSAMLWebEndpoint:
    """Test the handling of SAML Challenges."""

    @fixture
    def challenges_handled_event(self) -> asyncio.Event:
        """An asyncio Event to mark when all challenges are handled."""
        return asyncio.Event()

    @fixture
    def endpoint(
        self, challenges_handled_event: asyncio.Event, idp_url: str
    ) -> _SAMLWebEndpoint:
        """The SAML Web Endpoint to test against."""
        return _SAMLWebEndpoint(challenges_handled_event, idp_url)

    @fixture
    def aiohttp_app(self, endpoint: _SAMLWebEndpoint) -> Application:
        """Test AioHTTP application."""
        app = Application()
        app.add_routes([web.post("/api/saml", endpoint.handle_saml_idp_response)])
        return app

    @fixture
    async def client(
        self, aiohttp_app: Application, aiohttp_client: AiohttpClient
    ) -> TestClient:
        """Test client for AioHTTP interactions."""
        return await aiohttp_client(aiohttp_app)

    async def test_saml_responses_are_accumulated(
        self,
        aiohttp_app: Application,
        challenges_handled_event: asyncio.Event,
        client: TestClient,
        endpoint: _SAMLWebEndpoint,
        idp_url: str,
        mailbox_details: Dict[str, _WorkerMailboxDetails],
        saml_challenges: List[_DecryptedBitfountMessage],
        saml_responses: List[Dict[str, str]],
    ) -> None:
        """Test that we block and gather SAML responses from IdPs.

        This is testing that we don't shut down the server until
        all IdP responses have been received by the web server.
        This is achieved using an asyncio.Event().

        We also ensure that the web server redirects us to the
        next SAML URL.

        We also check that the saml responses are all passed
        to a function which will send them to the corresponding
        pods.
        """
        accumulated_saml_responses = []

        async def sender(
            saml_response: _SAMLResponse,
            _mailbox_details: _WorkerMailboxDetails,
        ) -> None:
            """Fake sending function.

            We use this to easily check that our SAML responses
            would be sent to the correct pods.
            """
            accumulated_saml_responses.append((saml_response, _mailbox_details))

        endpoint.set_saml_challenges(
            list(zip(saml_challenges, mailbox_details.values())), sender
        )

        # These requests would normally be made by the users browser
        # First SAML challenge Response
        first_redirect = await client.post(
            "/api/saml", data=saml_responses[0], allow_redirects=False
        )
        assert not challenges_handled_event.is_set()
        # The first redirect is for 2nd request,
        # as the first one would have been opened in the browser
        # The SAML response from the first challenge was sent in the POST above
        assert first_redirect.headers["location"] == f"{idp_url}saml-challenge-1"

        # Second SAML response
        second_redirect = await client.post(
            "/api/saml", data=saml_responses[1], allow_redirects=False
        )
        assert not challenges_handled_event.is_set()
        assert second_redirect.headers["location"] == f"{idp_url}saml-challenge-2"

        # Third SAML response
        success_page = await client.post("/api/saml", data=saml_responses[2])
        # The event should have occurred so that the web server can be shut down
        # And the modeller can go ahead with the rest of the flow
        assert challenges_handled_event.is_set()
        assert str(success_page.url.path) == "/api/saml"
        page_text = await success_page.text()
        assert (
            page_text == "You've now proven your identity to all pods involved"
            " in the task. You can close this tab."
        )

        await aiohttp_app.cleanup()

        # Check that send was called with the expected responses and mailboxes
        assert accumulated_saml_responses == list(
            zip(saml_responses, mailbox_details.values())
        )

    async def test_handle_throws_exception_when_not_configured(
        self, idp_url: str
    ) -> None:
        """Test exception thrown when send not set."""
        endpoint = _SAMLWebEndpoint(asyncio.Event(), idp_url)

        with pytest.raises(
            RuntimeError,
            match="SAML Challenges & Send method were not set "
            "in SAMLWebEndpoint. Unable to complete SAML"
            " authentication.",
        ):
            await endpoint.handle_saml_idp_response(Mock())


@unit_test
class TestSAMLChallengeHandler:
    """Test SAMLChallengeHandler."""

    @fixture
    def saml_challenge_handler(self, idp_url: str) -> _SAMLChallengeHandler:
        """Handler to use in tests."""
        return _SAMLChallengeHandler(idp_url)

    @fixture
    def mock_modeller_mailbox(
        self, mailbox_details: Dict[str, _WorkerMailboxDetails]
    ) -> Mock:
        """Mock _ModellerMailbox instance."""
        mock_mailbox: Mock = create_autospec(_ModellerMailbox, instance=True)
        mock_mailbox.worker_mailboxes = mailbox_details
        return mock_mailbox

    async def test_start_server_sets_server_start_task(
        self,
        saml_challenge_handler: _SAMLChallengeHandler,
    ) -> None:
        """Tests that the server start task is set.

        Nothing is mocked here, to ensure that the calls
        to aiohttp actually work!
        """
        saml_challenge_handler.start_server()

        assert saml_challenge_handler._server_start_task

    async def test__start_aiohttp_calls_work(
        self,
        saml_challenge_handler: _SAMLChallengeHandler,
    ) -> None:
        """Tests that the server start calls work.

        Nothing is mocked here, to ensure that the calls
        to aiohttp actually work!
        """
        await saml_challenge_handler._start()

    async def test__start_times_out_after_time(
        self,
        caplog: LogCaptureFixture,
        mocker: MockerFixture,
        saml_challenge_handler: _SAMLChallengeHandler,
    ) -> None:
        """Tests that _start() will timeout if server start hangs."""
        # Mock out runner
        mock_runner = mocker.patch.object(
            saml_challenge_handler, "runner", autospec=True
        )

        # Mock out site start-up method and init
        mock_tcp_site = create_autospec(TCPSite)
        mock_tcp_site_init = mocker.patch(
            "bitfount.federated.transport.identity_verification.saml.TCPSite",
            autospec=True,
            return_value=mock_tcp_site,
        )

        # Make wait_for raise timeout exception
        mocker.patch.object(
            asyncio, "wait_for", side_effect=TimeoutError("custom timeout")
        )

        with pytest.raises(TimeoutError, match="custom timeout"):
            await saml_challenge_handler._start()

        # Check setup called and site created
        mock_runner.setup.assert_awaited_once()
        mock_tcp_site_init.assert_called_once()

        # Check logged exception
        critical_logs = get_critical_logs(caplog)
        assert (
            "Timeout reached whilst trying to bind SAML web endpoint" in critical_logs
        )

    async def test__start_handles_address_in_use_error(
        self,
        caplog: LogCaptureFixture,
        mocker: MockerFixture,
        saml_challenge_handler: _SAMLChallengeHandler,
    ) -> None:
        """Tests that _start() handles address being unavailable."""
        # Mock out runner
        mock_runner = mocker.patch.object(
            saml_challenge_handler, "runner", autospec=True
        )

        # Mock out site start-up method and init
        mock_tcp_site = create_autospec(TCPSite)
        mock_tcp_site_init = mocker.patch(
            "bitfount.federated.transport.identity_verification.saml.TCPSite",
            autospec=True,
            return_value=mock_tcp_site,
        )
        # Make site.start() raise error indicating address unavailable
        mock_tcp_site.start.side_effect = OSError()

        with pytest.raises(OSError):
            await saml_challenge_handler._start()

        # Check setup called and site created
        mock_runner.setup.assert_awaited_once()
        mock_tcp_site_init.assert_called_once()

        # Check logged exception
        critical_logs = get_critical_logs(caplog)
        assert "Unable to bind SAML web endpoint" in critical_logs

    async def test_start_server_sets_up_server_in_background(
        self,
        mocker: MockerFixture,
        saml_challenge_handler: _SAMLChallengeHandler,
    ) -> None:
        """Ensure server is started in background with expected settings."""
        mock_tcp_site = mocker.patch(
            "bitfount.federated.transport.identity_verification.saml.TCPSite",
            autospec=True,
        )
        mock_create_task = mocker.patch(
            "bitfount.federated.transport.identity_verification.saml.create_task",
            wraps=create_task,
        )

        saml_challenge_handler.start_server()
        # Await start task to give time for background to run
        await saml_challenge_handler._server_start_task

        mock_create_task.assert_called_once()
        mock_tcp_site.assert_called_once_with(
            saml_challenge_handler.runner, "localhost", _BITFOUNT_MODELLER_PORT
        )
        mock_tcp_site.return_value.start.assert_awaited_once()

    async def test_saml_challenge_handler_opens_first_challenge(
        self,
        idp_url: str,
        mailbox_details: Dict[str, _WorkerMailboxDetails],
        mocker: MockerFixture,
        saml_challenge_handler: _SAMLChallengeHandler,
        saml_challenges: List[_DecryptedBitfountMessage],
    ) -> None:
        """Tests that the SAML authentication flow is triggered.

        This ensures that we set up the SAMLWebEndpoint as expected,
        and that we open a browser to begin the SAML auth flow.

        It also checks that we shut down our web server.
        """
        mock_endpoint = create_autospec(_SAMLWebEndpoint, instance=True)
        saml_challenge_handler.saml_endpoint = mock_endpoint

        challenges_handled = asyncio.Event()
        saml_challenge_handler.all_saml_challenges_handled = challenges_handled

        mock_webbrowser = mocker.patch(
            "bitfount.federated.transport.identity_verification.saml.webbrowser"
        )
        # Set event; mimicking the side effect of the listener
        mock_webbrowser.open.side_effect = lambda _: challenges_handled.set()

        # Mock the server start task
        saml_challenge_handler._server_start_task = AwaitableMock(spec=Task)
        saml_challenge_handler._server_start_task.done.return_value = False
        saml_challenge_handler.runner = create_autospec(AppRunner, instance=True)
        mock_send = Mock()

        await saml_challenge_handler._handle_saml_challenges(
            mock_send, saml_challenges, mailbox_details
        )

        # Ensure browser is opened to the first SAML URL
        mock_webbrowser.open.assert_called_once_with(
            f"{idp_url}{saml_challenges[0].body}"
        )
        # Ensure the web endpoint is given the SAML challenges
        mock_endpoint.set_saml_challenges.assert_called_once_with(
            list(zip(saml_challenges, mailbox_details.values())), mock_send
        )

        # Check that server task was awaited and interacted with
        saml_challenge_handler._server_start_task.done.assert_called_once()
        saml_challenge_handler._server_start_task.assert_awaited_once()
        saml_challenge_handler._server_start_task.result.assert_called_once()

    async def test_handle_method(
        self,
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        saml_challenge_handler: _SAMLChallengeHandler,
        saml_challenges: List[_DecryptedBitfountMessage],
    ) -> None:
        """Tests that the handle() method performs expected operations.

        In particular that it awaits the SAML challenges being sent and then
        passes these off to be handled.
        """
        # Mock out runner
        mock_runner = mocker.patch.object(
            saml_challenge_handler, "runner", autospec=True
        )

        # Create mock ModellerMailbox
        mock_modeller_mailbox.get_saml_challenges.return_value = saml_challenges

        # Mock out _handle_saml_challenges method
        mock__handle_saml_challenges = mocker.patch.object(
            saml_challenge_handler, "_handle_saml_challenges", autospec=True
        )

        await saml_challenge_handler.handle(mock_modeller_mailbox)

        # Check SAML challenges awaited
        mock_modeller_mailbox.get_saml_challenges.assert_awaited_once()
        # Check passed on to actual handler
        mock__handle_saml_challenges.assert_awaited_once_with(
            mock_modeller_mailbox.send_saml_responses,
            saml_challenges,
            mock_modeller_mailbox.worker_mailboxes,
        )
        # Check server shutdown at end
        mock_runner.cleanup.assert_awaited()

    async def test_handle_fails_if_server_not_started(
        self,
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        saml_challenge_handler: _SAMLChallengeHandler,
        saml_challenges: List[_DecryptedBitfountMessage],
    ) -> None:
        """Tests handle() fails if server not started."""
        # Mock out runner
        mock_runner = mocker.patch.object(
            saml_challenge_handler, "runner", autospec=True
        )

        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "SAML server has not been started; "
                "ensure start_server() has been called."
            ),
        ):
            await saml_challenge_handler.handle(mock_modeller_mailbox)

        # Check server is still shutdown
        mock_runner.cleanup.assert_awaited()

    @pytest.mark.parametrize(
        argnames="server_start_exception_cls",
        argvalues=(TimeoutError, OSError, Exception),
    )
    async def test_handle_reraises_exception_from_start_task(
        self,
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        saml_challenge_handler: _SAMLChallengeHandler,
        saml_challenges: List[_DecryptedBitfountMessage],
        server_start_exception_cls: Type[Exception],
    ) -> None:
        """Tests that server start exceptions are reraised when awaited."""
        # Mock server starting to make it throw an exception
        mocker.patch.object(
            saml_challenge_handler,
            "_start",
            autospec=True,
            side_effect=server_start_exception_cls("specific exception"),
        )
        saml_challenge_handler.start_server()
        assert saml_challenge_handler._server_start_task

        # Mock out runner and web endpoint
        mock_runner = mocker.patch.object(
            saml_challenge_handler, "runner", autospec=True
        )
        mocker.patch.object(saml_challenge_handler, "saml_endpoint", autospec=True)

        # Mock out SAML challenge reception
        mock_modeller_mailbox.get_saml_challenges.return_value = saml_challenges

        with pytest.raises(server_start_exception_cls):
            await saml_challenge_handler.handle(mock_modeller_mailbox)

        # Check server shutdown at end
        mock_runner.cleanup.assert_awaited()



"""Useful types for Federated Learning."""
from __future__ import annotations

from enum import Enum, auto
from typing import TYPE_CHECKING, List, Optional, Protocol, TypedDict, runtime_checkable

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from typing_extensions import NotRequired

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub


class SerializedModel(TypedDict):
    """Serialized representation of a model."""

    class_name: str
    hub: NotRequired[Optional[BitfountHub]]


class SerializedAlgorithm(TypedDict):
    """Serialized representation of an algorithm."""

    class_name: str  # value from AlgorithmType enum
    model: NotRequired[SerializedModel]


class SerializedAggregator(TypedDict):
    """Serialized representation of an aggregator."""

    class_name: str  # value from AggregatorType enum


class SerializedProtocol(TypedDict):
    """Serialized representation of a protocol."""

    class_name: str  # value from ProtocolType enum
    algorithm: SerializedAlgorithm
    aggregator: NotRequired[SerializedAggregator]


class ProtocolType(Enum):
    """Available protocol names from `bitfount.federated.protocol`."""

    FederatedAveraging = "bitfount.FederatedAveraging"
    ResultsOnly = "bitfount.ResultsOnly"
    PrivateSetIntersection = "bitfount.PrivateSetIntersection"


class AlgorithmType(Enum):
    """Available algorithm names from `bitfount.federated.algorithm`."""

    FederatedModelTraining = "bitfount.FederatedModelTraining"
    ModelTrainingAndEvaluation = "bitfount.ModelTrainingAndEvaluation"
    ModelEvaluation = "bitfount.ModelEvaluation"
    ColumnAverage = "bitfount.ColumnAverage"
    SqlQuery = "bitfount.SqlQuery"
    PrivateSqlQuery = "bitfount.PrivateSqlQuery"
    ComputeIntersectionRSA = "bitfount.ComputeIntersectionRSA"


class AggregatorType(Enum):
    """Available aggregator names from `bitfount.federated.aggregator`."""

    Aggregator = "bitfount.Aggregator"
    SecureAggregator = "bitfount.SecureAggregator"


class _PodResponseType(Enum):
    """Pod response types sent to `Modeller` on a training job request."""

    # Common response types
    ACCEPT = auto()

    # /api/access response types
    NO_ACCESS = auto()
    INVALID_PROOF_OF_IDENTITY = auto()
    UNAUTHORISED = auto()
    NO_PROOF_OF_IDENTITY = auto()


class _DataLessAlgorithm:
    """Base algorithm class for tagging purposes.

    Used in algorithms for which data loading is done at runtime.
    """

    ...


_RESPONSE_MESSAGES = {
    # Common response messages
    _PodResponseType.ACCEPT: "Job accepted",
    # /api/access response messages
    _PodResponseType.NO_ACCESS: "There are no permissions for this modeller/pod combination.",  # noqa: B950
    _PodResponseType.INVALID_PROOF_OF_IDENTITY: "Unable to verify identity; ensure correct login used.",  # noqa: B950
    _PodResponseType.UNAUTHORISED: "Insufficient permissions for the requested task on this pod.",  # noqa: B950
    _PodResponseType.NO_PROOF_OF_IDENTITY: "Unable to verify identity, please try again.",  # noqa: B950
}


@runtime_checkable
class _TaskRequestMessageGenerator(Protocol):
    """Callback protocol describing a task request message generator."""

    def __call__(
        self,
        serialized_protocol: SerializedProtocol,
        pod_identifiers: List[str],
        aes_key: bytes,
        pod_public_key: RSAPublicKey,
    ) -> bytes:
        """Function signature for the callback."""
        ...



"""Tests dataloaders.py."""
import numpy as np
import pandas as pd
import pytest
from pytest import fixture

from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasets import _BitfountDataset
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datasplitters import PercentageSplitter
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import DataSplit, SemanticType
from tests.utils.helper import TABLE_NAME, create_dataset, unit_test


class TestDataLoaders:
    """Tests BitfountDataloader."""

    @fixture
    def dataframe(self) -> pd.DataFrame:
        """Dataframe fixture."""
        df = create_dataset(image=True)
        # Drop the date column until we support datetime.
        return df.drop(columns=["Date"])

    @unit_test
    def test_get_x_dataframe_tabular_only(self, dataframe: pd.DataFrame) -> None:
        """Tests get_x_dataframe for tabular data."""
        df = dataframe.drop(columns=["image"])
        datasource = DataFrameSource(df)
        datasource.load_data()
        datastucture = DataStructure(target="TARGET", table=TABLE_NAME)
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datastucture.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _BitfountDataset(
            datasource=datasource,
            target="TARGET",
            selected_cols_semantic_types=datastucture.selected_cols_w_types,
            selected_cols=datastucture.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        dl = BitfountDataLoader(dataset=dataset)
        x_df = dl.get_x_dataframe()
        # Drop target and text columns
        new_df = df.drop(columns=["TARGET", "I", "J", "K", "L"])
        assert isinstance(x_df, pd.DataFrame)
        assert set(x_df.columns) == set(new_df.columns)

    @unit_test
    def test_get_x_dataframe_image_only(self, dataframe: pd.DataFrame) -> None:
        """Tests get_x_dataframe for tabular data."""
        df = dataframe[["image", "TARGET"]]
        datasource = DataFrameSource(df, data_splitter=PercentageSplitter(0, 0))
        datasource.load_data()
        datastucture = DataStructure(
            target="TARGET", table=TABLE_NAME, selected_cols=["image", "TARGET"]
        )
        schema = BitfountSchema()
        schema.add_datasource_tables(
            datasource,
            force_stypes={TABLE_NAME: {"image": ["image"]}},
            table_name=TABLE_NAME,
        )
        datastucture.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _BitfountDataset(
            datasource=datasource,
            target="TARGET",
            selected_cols_semantic_types=datastucture.selected_cols_w_types,
            selected_cols=datastucture.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        dl = BitfountDataLoader(dataset=dataset)
        x_df = dl.get_x_dataframe()
        assert isinstance(x_df, pd.DataFrame)
        assert x_df.shape == df[["image"]].shape
        assert x_df.columns == df[["image"]].columns

    @unit_test
    def test_get_x_dataframe_image_and_tab(self, dataframe: pd.DataFrame) -> None:
        """Tests get_x_dataframe for tabular data."""
        datasource = DataFrameSource(dataframe, data_splitter=PercentageSplitter(0, 0))
        datasource.load_data()
        datastucture = DataStructure(target="TARGET", table=TABLE_NAME)
        schema = BitfountSchema()
        schema.add_datasource_tables(
            datasource,
            force_stypes={TABLE_NAME: {"image": ["image"]}},
            table_name=TABLE_NAME,
        )
        datastucture.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _BitfountDataset(
            datasource=datasource,
            target="TARGET",
            selected_cols_semantic_types=datastucture.selected_cols_w_types,
            selected_cols=datastucture.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        dl = BitfountDataLoader(dataset=dataset)
        x_df = dl.get_x_dataframe()
        # Drop target, image, and text columns
        new_df = dataframe.drop(columns=["TARGET", "I", "J", "K", "L", "image"])
        # x_df is a tuple of dataframes (tabular followed by image)
        assert isinstance(x_df, tuple)
        assert set(x_df[0].columns) == set(new_df.columns)
        assert isinstance(x_df[1], pd.DataFrame)
        assert x_df[1].shape == dataframe[["image"]].shape
        assert x_df[1].columns == dataframe[["image"]].columns

    @unit_test
    def test_empty_dataframe_raises_valerror(self, dataframe: pd.DataFrame) -> None:
        """Tests get_x_dataframe with empty df raises error."""
        df = pd.DataFrame(dataframe[["TARGET"]])
        datasource = DataFrameSource(df)
        datasource.load_data()
        datastucture = DataStructure(
            target="TARGET", selected_cols=["TARGET"], table=TABLE_NAME
        )
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datastucture.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _BitfountDataset(
            datasource=datasource,
            target="TARGET",
            selected_cols_semantic_types=datastucture.selected_cols_w_types,
            selected_cols=datastucture.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        dl = BitfountDataLoader(dataset=dataset)
        with pytest.raises(ValueError):
            dl.get_x_dataframe()

    @unit_test
    def test_dataloader_tab(self, tabular_dataset: _BitfountDataset) -> None:
        """Tests x- and y-dataframe retrieval from dataloader for tabular data."""
        df = BitfountDataLoader(tabular_dataset)
        x = df.get_x_dataframe()
        y = df.get_y_dataframe()
        assert isinstance(x, pd.DataFrame)
        assert len(x) == len(y)
        assert len(x.columns) == 12
        assert len(y.columns) == 1

    @unit_test
    def test_dataloader_img(self, image_dataset: _BitfountDataset) -> None:
        """Tests x- and y-dataframe retrieval from dataloader for image data."""
        df = BitfountDataLoader(image_dataset)
        x = df.get_x_dataframe()
        y = df.get_y_dataframe()
        assert isinstance(x, pd.DataFrame)
        assert len(x) == len(y)
        assert len(x.columns) == 1
        assert len(y.columns) == 1

    @unit_test
    def test_dataloader_img_tab(self, image_tab_dataset: _BitfountDataset) -> None:
        """Tests x- and y-dataframe retrieval from dataloader."""
        df = BitfountDataLoader(image_tab_dataset)
        x = df.get_x_dataframe()
        y = df.get_y_dataframe()
        assert isinstance(x, tuple)
        tab, img = x
        assert len(tab) == len(y)
        assert len(img) == len(y)
        assert len(y.columns) == 1
        assert len(tab.columns) == 12
        assert len(img.columns) == 1

    @unit_test
    def test_dataloader_multiimage(self, multiimage_dataset: _BitfountDataset) -> None:
        """Tests x- and y-dataframe retrieval from dataloader for multi-image."""
        df = BitfountDataLoader(multiimage_dataset)
        x = df.get_x_dataframe()
        y = df.get_y_dataframe()
        assert isinstance(x, pd.DataFrame)
        assert len(x) == len(y)
        assert len(x.columns) == 2
        assert len(y.columns) == 1

    @unit_test
    def test_dataloader_multilabel(self) -> None:
        """Tests x- and y-dataframe retrieval from dataloader for multilabel target."""
        data = create_dataset(image=True)
        data = data.assign(TARGET_2=np.zeros(len(data)))
        data.loc[(data.A < 700) & (data.F < 0.5) & (data.D % 2 == 1), "TARGET_2"] = 1
        datasource = DataFrameSource(data)
        datasource.load_data()
        schema = BitfountSchema()
        schema.add_datasource_tables(
            datasource,
            force_stypes={TABLE_NAME: {"image": ["image"]}},
            table_name=TABLE_NAME,
        )
        datasource.data = datasource.data.drop(
            columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
        )
        datastructure = DataStructure(target=["TARGET", "TARGET_2"], table=TABLE_NAME)
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        multilabel_dataset = _BitfountDataset(
            datasource=datasource,
            target=["TARGET", "TARGET_2"],
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        df = BitfountDataLoader(multilabel_dataset)
        x = df.get_x_dataframe()
        y = df.get_y_dataframe()
        assert isinstance(x, tuple)
        tab, img = x
        assert len(tab) == len(y)
        assert len(img) == len(y)
        assert len(y.columns) == 2
        assert len(tab.columns) == 13
        assert len(img.columns) == 1



"""Tests for the transformation parsing."""
from pathlib import Path
from typing import Dict, Tuple

import pytest
from pytest import fixture
import yaml

from bitfount.data.types import DataSplit
from bitfount.transformations.batch_operations import AlbumentationsImageTransformation
from bitfount.transformations.binary_operations import (
    AdditionTransformation,
    DivisionTransformation,
)
from bitfount.transformations.exceptions import TransformationParsingError
from bitfount.transformations.parser import TransformationsParser
from bitfount.transformations.unary_operations import (
    InclusionTransformation,
    OneHotEncodingTransformation,
)
from tests.utils.helper import unit_test

test_yaml = """
Transformations:
- Add:
    name: col_1_and_2_sum
    output: false
    arg1: col:col_1
    arg2: c:col_2
- Divide:
    name: normalised_result
    output: true
    arg1: tran:col_1_and_2_sum
    arg2: 100
- In:
    in_str: "DAILY"
    arg: column:frequency
    output: true
"""

out_col_refs = {"col_1", "col_2", "frequency"}


@fixture
def test_transformations(
    fake_uuid: None,
) -> Tuple[AdditionTransformation, DivisionTransformation, InclusionTransformation]:
    """Transformations for tests."""
    return (
        AdditionTransformation(
            name="col_1_and_2_sum", output=False, arg1="col:col_1", arg2="c:col_2"
        ),
        DivisionTransformation(
            name="normalised_result", output=True, arg1="tran:col_1_and_2_sum", arg2=100
        ),
        InclusionTransformation(output=True, in_str="DAILY", arg="column:frequency"),
    )


@fixture
def test_transformations_results(
    test_transformations: Tuple[
        AdditionTransformation, DivisionTransformation, InclusionTransformation
    ]
) -> Tuple[AdditionTransformation, DivisionTransformation, InclusionTransformation]:
    """Generate transformations where transformations are hooked together."""
    test_transformations[1].arg1 = test_transformations[0]  # type: ignore[assignment] # Reason: replacing str with Transformation is expected # noqa: B950
    return test_transformations


@fixture
def transformation_parser() -> TransformationsParser:
    """A TransformationsParser instance for tests."""
    return TransformationsParser()


@unit_test
class TestTransformationsParser:
    """Tests for the TransformationsParser class."""

    def test_parse(
        self,
        fake_uuid: None,
        test_transformations_results: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests parsing of transformations string works correctly."""
        transformations, col_refs = transformation_parser.parse(test_yaml)
        assert transformations == list(test_transformations_results)
        assert col_refs == out_col_refs

    def test_parse_file(
        self,
        fake_uuid: None,
        test_transformations_results: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        tmp_path: Path,
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests parse file works correctly."""
        test_file = tmp_path / "test.yml"
        with open(test_file, "w") as f:
            f.write(test_yaml)
        transformations, col_refs = transformation_parser.parse_file(test_file)
        assert transformations == list(test_transformations_results)
        assert col_refs == out_col_refs

    def test_deserialize_transformations(
        self,
        test_transformations_results: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests deserialize_transformations works correctly."""
        data = yaml.safe_load(test_yaml)["Transformations"]
        transformations, col_refs = transformation_parser.deserialize_transformations(
            data
        )
        assert transformations == list(test_transformations_results)
        assert col_refs == out_col_refs

    def test_deserialize_transformations_captures_errors(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests that all errors are captured and output at once."""
        bad_yaml = """
        Transformations:
        - UnknownTransform:
            name: unknown_1
        - Add:
            name: multi_transform_1
            arg1: test
            arg2: 100
          Divide:
            name: multi_transform_2
            arg1: test
            arg2: 100
        - Add:
            name: normalised_result
            arg1: test
            arg2: 100
        - Divide:
            name: normalised_result
            output: true
            arg1: tran:col_1_and_2_sum
            arg2: 100
        - In:
            in_str: "DAILY"
            arg: column:frequency
            output: true
        - OneHotEncode:
            name: normalised
            arg: column:is_normalised
            values: [result]
        """
        data = yaml.safe_load(bad_yaml)["Transformations"]
        with pytest.raises(TransformationParsingError) as tpe:
            transformation_parser.deserialize_transformations(data)
        assert tpe.value.errors == [
            f'No transformation registered with name {"UnknownTransform".lower()}.',
            "Each transformation mapping must contain exactly one transformation; "
            "mapping 1 contains 2.",
            "Duplicate transformation name: normalised_result.",
            "Multi-column output clash: normalised_result "
            "(from output column of normalised).",
            "Transformation, normalised_result, attempted to use transformation "
            '"col_1_and_2_sum" before it was defined.',
        ]

    def test__deserialize_transformation(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests _deserialize_transformation works correctly."""
        t_spec = {"Add": {"name": "add_test", "arg1": "arg1_test", "arg2": "arg2_test"}}
        t = transformation_parser._deserialize_transformation(0, t_spec)
        assert t == AdditionTransformation(
            name="add_test", arg1="arg1_test", arg2="arg2_test"
        )

    def test__deserialize_transformation_fails_on_multi_spec(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests deserialization fails if multiple specs in one transformation."""
        t_spec: Dict[str, dict] = {"Add": {}, "Divide": {}}
        with pytest.raises(TransformationParsingError):
            transformation_parser._deserialize_transformation(0, t_spec)

    def test__deserialize_transformation_fails_on_empty_spec(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests deserialization fails if a transformation spec is empty."""
        t_spec: Dict[str, dict] = {}
        with pytest.raises(TransformationParsingError):
            transformation_parser._deserialize_transformation(0, t_spec)

    def test__deserialize_transformation_fails_on_non_registered_transform(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests deserialization raises exception on unknown transformation."""
        t_spec: Dict[str, dict] = {"NotARealTransform": {}}
        with pytest.raises(TransformationParsingError):
            transformation_parser._deserialize_transformation(0, t_spec)

    def test__check_names_unique(
        self,
        test_transformations: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests _check_names_unique works correctly."""
        transformation_parser._check_names_unique(test_transformations)

    def test__check_names_unique_w_multi_column(
        self,
        test_transformations: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests _check_names_unique works correctly with multi-output columns."""
        aux_test_transformations = test_transformations + (
            OneHotEncodingTransformation(
                name="one_hot", arg="c:categorical", raw_values=["CAT_1", "CAT_2"]
            ),
        )
        transformation_parser._check_names_unique(aux_test_transformations)

    def test__check_names_unique_fails(
        self,
        test_transformations: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests _check_names_unique fails on duplicate names."""
        test_transformations[2].name = test_transformations[0].name
        test_transformations[1].name = test_transformations[0].name

        with pytest.raises(TransformationParsingError) as tpe:
            transformation_parser._check_names_unique(test_transformations)

        assert tpe.value.errors == [
            f"Duplicate transformation name: {test_transformations[0].name}."
        ]

    def test__check_names_unique_fails_w_multi_column_same_name(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests _check_names_unique fails on duplicate name in MCO."""
        test_transformations = [
            AdditionTransformation(
                name="col_1_and_2_sum", output=False, arg1="col:col_1", arg2="c:col_2"
            ),
            OneHotEncodingTransformation(
                name="col_1_and_2_sum", arg="c:categorical", raw_values=["_unused"]
            ),
        ]

        with pytest.raises(TransformationParsingError) as tpe:
            transformation_parser._check_names_unique(test_transformations)

        assert tpe.value.errors == ["Duplicate transformation name: col_1_and_2_sum."]

    def test__check_names_unique_fails_w_multi_column_clash(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests _check_names_unique fails in multi-column scenario.

        Fails if the output columns of a multi-column transformation output would
        clash with another column name.
        """
        test_transformations = [
            AdditionTransformation(
                name="col_1_and_2_sum", output=False, arg1="col:col_1", arg2="c:col_2"
            ),
            OneHotEncodingTransformation(
                name="col_1", arg="c:categorical", raw_values=["and_2_sum"]
            ),
        ]

        with pytest.raises(TransformationParsingError) as tpe:
            transformation_parser._check_names_unique(test_transformations)

        assert tpe.value.errors == [
            "Multi-column output clash: col_1_and_2_sum (from output column of col_1)."
        ]

    def test__check_names_unique_fails_w_multiple_multi_column_clash(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests _check_names_unique fails in multiple multi-column scenario.

        Fails if the output columns of a multi-column transformation output would
        clash with the output column(s) of another multi-column transformation.
        """
        # Test correctly fails if the second MCO clashes with the first,
        # but the first doesn't clash with a base.
        test_transformations = [
            AdditionTransformation(
                name="col_1_and_2_sum", output=False, arg1="col:col_1", arg2="c:col_2"
            ),
            OneHotEncodingTransformation(
                name="col_1", arg="c:categorical", raw_values=["not_a_clash"]
            ),
            OneHotEncodingTransformation(
                name="col_1_not_a", arg="c:categorical", raw_values=["clash"]
            ),
        ]

        with pytest.raises(TransformationParsingError) as tpe:
            transformation_parser._check_names_unique(test_transformations)

        assert tpe.value.errors == [
            "Multi-column output clash: col_1_not_a_clash "
            "(from output column of col_1_not_a)."
        ]

        # Test correctly fails if the second MCO clashes with the first,
        # and the first also clashes with a base.
        test_transformations = [
            AdditionTransformation(
                name="col_1_and_2_sum", output=False, arg1="col:col_1", arg2="c:col_2"
            ),
            OneHotEncodingTransformation(
                name="col_1", arg="c:categorical", raw_values=["and_2_sum"]
            ),
            OneHotEncodingTransformation(
                name="col_1_and", arg="c:categorical", raw_values=["2_sum"]
            ),
        ]

        with pytest.raises(TransformationParsingError) as tpe:
            transformation_parser._check_names_unique(test_transformations)

        assert tpe.value.errors == [
            "Multi-column output clash: col_1_and_2_sum "
            "(from output column of col_1).",
            "Multi-column output clash: col_1_and_2_sum "
            "(from output column of col_1_and).",
        ]

    def test__hook_transformations_together(
        self,
        fake_uuid: None,
        test_transformations: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests _hook_transformations_together works correctly."""
        transformation_parser._hook_transformations_together(test_transformations)

        # First and last transformations unchanged
        assert test_transformations[0] == AdditionTransformation(
            name="col_1_and_2_sum", output=False, arg1="col:col_1", arg2="c:col_2"
        )
        assert test_transformations[2] == InclusionTransformation(
            output=True, in_str="DAILY", arg="column:frequency"
        )

        # Second transformation now references first, otherwise same
        orig_second = DivisionTransformation(
            name="normalised_result", output=True, arg1="tran:col_1_and_2_sum", arg2=100
        )
        assert test_transformations[1].name == orig_second.name
        assert test_transformations[1].output == orig_second.output
        assert test_transformations[1].arg2 == orig_second.arg2
        assert test_transformations[1].arg1 == test_transformations[0]

    def test__hook_transformations_together_fails_forward_ref(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests that parsing fails with exception if cols are forward referenced."""
        transformations = [
            AdditionTransformation(name="add", arg1="t:test", arg2=""),
            DivisionTransformation(name="div", arg1="tran:test", arg2=""),
            InclusionTransformation(name="test", in_str="", arg=""),
        ]

        with pytest.raises(TransformationParsingError) as tpe:
            transformation_parser._hook_transformations_together(transformations)

        assert tpe.value.errors == [
            f'Transformation, {name}, attempted to use transformation "{ref}" '
            f"before it was defined."
            for name, ref in [("add", "test"), ("div", "test")]
        ]

    def test__extract_column_refs(
        self,
        test_transformations: Tuple[
            AdditionTransformation, DivisionTransformation, InclusionTransformation
        ],
        transformation_parser: TransformationsParser,
    ) -> None:
        """Tests that _extract_column_refs() works correctly."""
        col_refs = transformation_parser._extract_column_refs(test_transformations)
        assert col_refs == out_col_refs


@unit_test
class TestTransformationParsingError:
    """Tests for the TransformationParsingError exception."""

    def test_single_error_arg(self) -> None:
        """Tests creation of TransformationParsingError with a single error."""
        with pytest.raises(TransformationParsingError) as tpe:
            raise TransformationParsingError("test")
        assert tpe.value.errors == ["test"]

    def test_multiple_error_arg(self) -> None:
        """Tests creation of TransformationParsingError with multiple errors."""
        with pytest.raises(TransformationParsingError) as tpe:
            raise TransformationParsingError(["test", "test2"])
        assert tpe.value.errors == ["test", "test2"]

    def test_str(self) -> None:
        """Tests string representation of TransformationParsingError."""
        with pytest.raises(TransformationParsingError) as tpe:
            raise TransformationParsingError(["test", "test2"])
        assert str(tpe.value) == "Errors: \ntest\ntest2"


@unit_test
class TestParsingBatchTransformations:
    """Tests parsing of batch transformations."""

    def test_parsing_image_transformation_successfully(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests parsing of image transformation succesfully."""
        transformations_yaml = """
        Transformations:
            - albumentations:
                arg: col:col_1
                name: test_col
                step: train
                output: True
                transformations:
                    - ResizeTransformation:
                        arg1: 1
                        arg2: 2
                    - NormalizeTransformation
        """
        tfms, _ = transformation_parser.parse(transformations_yaml)
        assert tfms[0] == AlbumentationsImageTransformation(
            name="test_col",
            output=True,
            arg="col:col_1",
            step=DataSplit.TRAIN,
            transformations=[
                {"ResizeTransformation": {"arg1": 1, "arg2": 2}},
                "NormalizeTransformation",
            ],
        )

    def test_image_transformation_output_must_be_true(
        self, transformation_parser: TransformationsParser
    ) -> None:
        """Tests that output must be True for Image Transformation."""
        transformations_yaml = """
        Transformations:
            - albumentations:
                arg: col:col_1
                name: test_col
                step: train
                output: False
                transformations:
                    - ResizeTransformation:
                        arg1: 1
                        arg2: 2
                    - NormalizeTransformation
        """
        with pytest.raises(
            ValueError, match="`output` cannot be False for a BatchTimeOperation"
        ):
            transformation_parser.parse(transformations_yaml)



"""Secure aggregation."""
from __future__ import annotations

import inspect
import logging
import secrets
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Mapping, Type, Union

from marshmallow import fields
import numpy as np

from bitfount.federated.shim import _load_default_tensor_shim
from bitfount.federated.transport.worker_transport import (
    _get_worker_secure_shares,
    _InterPodWorkerMailbox,
    _send_secure_shares_to_others,
)
from bitfount.types import (
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    _BaseSerializableObjectMixIn,
    _Weights,
)

if TYPE_CHECKING:
    from bitfount.types import _TensorLike

# Can't be larger than 2^64 -1 (largest unsigned 64 bit integer). Otherwise we get:
# “OverflowError: Python int too large to convert to C long"
LARGE_PRIME_NUMBER: int = (2**61) - 1  # Largest possible Mersenne prime number.

# Precision does not need to be greater than this in order be able to perform lossless
# computation on IEEE 754 32-bit floating point values
FLOAT_32_BIT_PRECISION: int = 10**10

_secure_share_registry: Dict[str, Type[_BaseSecureShare]] = {}
secure_share_registry: Mapping[str, Type[_BaseSecureShare]] = MappingProxyType(
    _secure_share_registry
)
logger = logging.getLogger(__name__)


class _BaseSecureShare:
    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
            _secure_share_registry[cls.__name__] = cls


class SecureShare(_BaseSecureShare, _BaseSerializableObjectMixIn):
    """Additive, replicated, secret sharing algorithm responsible for secure averaging.

    This secret sharing implementation is 'additive' because the secret can be
    reconstructed by taking the sum of all the shares it is split into, and 'replicated'
    because each party receives more than one share.

    The algorithm works as follows:
        1. First every worker shares a securely generated random number (between 0 and
        `prime_q`) with every other worker such that every worker ends up with one
        number from every other worker. These numbers are known as shares as they will
        form part of the secret (the state dictionary) which will be shared.
        2. The values in the state dictionary are then converted to positive integer
        field elements of a finite field bounded by `prime_q`.
        3. The random numbers generated are used to compute a final share for every
        value in the state dictionary. This final share has the same shape as the secret
        state dictionary.
        4. This final share is then reconstructed using the shares retrieved from the
        other workers. At this point, the final share from each worker is meaningless
        until averaged with every other state dictionary.
        5. This final share is sent to the modeller where it will be averaged with the
        state dictionaries from all the other workers (all the while in the finite field
        space).
        6. After averaging, the state dictionaries are finally decoded back to floating
        point numpy arrays.

    :::note

    The relationships between individual elements in the tensors are preserved in
    this implementation since our shares are scalars rather than vectors. Therefore,
    whilst the secret itself cannot be reconstructed, some properties of the secret can
    be deciphered e.g. which element is the largest/smallest, etc.

    :::

    Args:
        prime_q: Large prime number used in secure aggregation. This should be a
            few orders of magnitude larger than the precision so that when we add
            encoded finite field elements with one another, we do not breach the limits
            of the finite field. A `SecureShareError` is raised if this occurs. Defaults
            to 2^61 -1 (the largest Mersenne 64 bit Mersenne prime number - for ease).
        precision: Degree of precision for floating points in secure aggregation
            i.e. the number of digits after the decimal point that we want to keep.
            Defaults to 10^10.

    Attributes:
        prime_q: Large prime number used in secure aggregation.
        precision: Degree of precision for floating points in secure aggregation.
    """

    # TODO: [BIT-423] Review security
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "prime_q": fields.Integer(),
        "precision": fields.Integer(),
    }
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}

    def __init__(
        self,
        prime_q: int = LARGE_PRIME_NUMBER,
        precision: int = FLOAT_32_BIT_PRECISION,
    ):
        self.class_name = type(self).__name__
        self.prime_q = prime_q
        self.precision = precision

        self._tensor_shim = _load_default_tensor_shim()
        self._own_shares: List[int] = []
        self._other_worker_shares: List[int] = []

    def _encode_finite_field(
        self, rational: Union[_TensorLike, np.ndarray]
    ) -> np.ndarray:
        """Converts `rational` to integer in finite field.

        Raises:
            SecureShareError: if finite field limit is breached. This is raised if there
                are not enough integers to represent all the possible floating point
                numbers in `rational`.
        """
        total_num_workers = len(self._own_shares) + 1
        # Warning! on Windows astype(int) returns the min long value.
        # We use astype(np.int64) in this module to avoid this.

        # This branch is only for tensors. Clamping parameters does not apply to
        # non-model state dictionaries.
        if not isinstance(rational, np.ndarray):
            # Auxiliary value for checking finite field limit
            _upscaled_param_values = (
                self._tensor_shim.to_numpy(rational * self.precision).astype(np.int64)
                * total_num_workers
            )

            # Check that _upscaled_param_values are all within finite field limits
            if (
                (_upscaled_param_values > self.prime_q / 2)
                | (_upscaled_param_values < -self.prime_q / 2)
            ).sum() != 0:
                logger.warning(
                    "Parameter weights have been clipped. If you want to avoid this, "
                    "choose a larger `prime_q` or a smaller `precision` for "
                    "the `SecureShare` or normalize continuous features prior "
                    "to training."
                )
                rational = self._tensor_shim.clamp_params(
                    rational, self.prime_q, self.precision, total_num_workers
                )

            rational = self._tensor_shim.to_numpy(rational)

        upscaled = (rational * self.precision).astype(np.int64)
        field_element = upscaled % self.prime_q
        return field_element

    def _decode_finite_field(self, field_element: np.ndarray) -> np.ndarray:
        """Converts finite field array back into a rational array."""
        field_element = np.where(
            field_element > (self.prime_q / 2),
            field_element - self.prime_q,
            field_element,
        )
        return (field_element / self.precision).astype(float)

    def _encode_secret(self, secret: Union[_TensorLike, np.ndarray]) -> np.ndarray:
        """Encodes the provided secret using `self.own_shares` and returns it.

        Secret is first moved to the finite field space and then split into n shares
        where n is the number of all workers participating in training. All but one
        shares (integers) are shared with the other workers and the final share
        (a dictionary of tensors) is returned. The sum of all these will yield the
        original (encoded) secret.
        """
        secret_array = self._encode_finite_field(secret)
        encoded_secret = (secret_array - sum(self._own_shares)) % self.prime_q
        return encoded_secret

    def _reconstruct_secret(self, shares: List[Union[np.ndarray, int]]) -> np.ndarray:
        """Reconstructs the shares into a secret.

        This secret is not the same secret as originally encoded and shared. This secret
        is useless unless averaged with the secret outputs from all the other workers.
        """
        return np.asarray(sum(shares) % self.prime_q)

    def _encode_and_reconstruct_state_dict(
        self, secret_state_dict: Union[_Weights, Mapping[str, np.ndarray]]
    ) -> Dict[str, np.ndarray]:
        """Encodes and reconstructs state dict from own and worker shares.

        Encrypts `secret_state_dict` using `own_shares` then reconstructs it using
        `self._other_worker_shares`.
        """
        logger.debug("Encoding state dict...")
        encrypted_state_dict = [
            self._encode_secret(v) for _, v in secret_state_dict.items()
        ]
        logger.debug("Reconstructing state dict...")
        reconstructed = [
            self._reconstruct_secret([*self._other_worker_shares, value])
            for value in encrypted_state_dict
        ]
        return dict(zip(list(secret_state_dict), reconstructed))

    def _get_random_number(self) -> int:
        """Generate a random number, append it to `self.own_shares` and also return it.

        Random number generator is cryptographically secure.
        """
        rand_num = secrets.randbelow(self.prime_q)
        self._own_shares.append(rand_num)
        return rand_num

    async def _share_own_shares(self, mailbox: _InterPodWorkerMailbox) -> None:
        """Sends own secure aggregation shares to other workers (one each).

        A random number is securely generated and sent to each worker such that each
        worker receives a different random number from every other worker. Each worker
        keeps a copy of the random numbers they generated which later become 'shares'
        as they can be used to encode a secret (the state dictionary).
        """
        self._own_shares = []
        await _send_secure_shares_to_others(self._get_random_number, mailbox)

    async def _receive_worker_shares(self, mailbox: _InterPodWorkerMailbox) -> None:
        """Receives secure aggregation shares from other workers."""
        self._other_worker_shares = await _get_worker_secure_shares(mailbox)

    def _add(self, arrays: List[np.ndarray]) -> np.ndarray:
        """Add multiple encoded numpy arrays element-wise and return a numpy array.

        All arrays must have the same shape.
        """
        if arrays[0].size > 1:
            return np.array(
                [np.sum(arrs, axis=0) % self.prime_q for arrs in zip(*arrays)]
            )
        else:
            return np.asarray(np.sum(arrays, axis=0) % self.prime_q)

    def average_and_decode_state_dicts(
        self,
        state_dicts: List[Dict[str, np.ndarray]],
    ) -> Dict[str, np.ndarray]:
        """Averages and decodes multiple encrypted state dictionaries.

        Computes the mean of all the `state_dicts` before decoding the averaged result
        and returning it. This is called on the Modeller side after receiving the state
        dictionaries from all the workers.

        Args:
            state_dicts: List of encoded state dictionaries as numpy arrays.

        Returns:
            A dictionary of averaged and decoded state dictionaries.
        """
        average_state_dict: Dict[str, np.ndarray] = {}
        for field in state_dicts[0]:
            summed_value = self._add([state_dict[field] for state_dict in state_dicts])
            average_decoded_value = self._decode_finite_field(summed_value) / len(
                state_dicts
            )
            average_state_dict[field] = average_decoded_value

        return average_state_dict

    async def do_secure_aggregation(
        self,
        state_dict: Union[_Weights, Mapping[str, np.ndarray]],
        mailbox: _InterPodWorkerMailbox,
    ) -> Dict[str, np.ndarray]:
        """Performs secure aggregation.

        This is called on the Pod side before sending the state dict to the
        Modeller.

        Args:
            state_dict: A dictionary of tensors or numpy arrays to be securely
                aggregated.
            mailbox: A mailbox to send and receive messages from other workers.

        Returns:
            A dictionary of encoded numpy arrays.

        Raises:
            SecureShareError: If finite field limit is breached. This is raised if there
                are not enough integers to represent all the possible floating point
                numbers.
        """
        logger.info("Performing secure aggregation...")
        await self._share_own_shares(mailbox)
        logger.debug("Shared own shares with other workers.")
        await self._receive_worker_shares(mailbox)
        logger.debug("Received shares from other workers.")
        encoded_reconstructued_update = self._encode_and_reconstruct_state_dict(
            state_dict
        )
        logger.info("Secure aggregation complete.")
        return encoded_reconstructued_update



"""Fixtures for authorisation-checkers-related classes."""
from dataclasses import dataclass
from typing import cast
from unittest.mock import Mock, NonCallableMock, create_autospec

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from pytest import fixture

from bitfount.federated.authorisation_checkers import (
    _IDENTITY_VERIFICATION_METHODS_MAP,
    IdentityVerificationMethod,
    _SignatureBasedAuthorisation,
)
from bitfount.federated.types import _TaskRequestMessageGenerator


@fixture
def default_task_request_msg_gen() -> _TaskRequestMessageGenerator:
    """Creates a TaskRequestMessageGenerator for the default verification method."""
    # Create default task_request_msg_gen
    default_authoriser = _IDENTITY_VERIFICATION_METHODS_MAP[
        IdentityVerificationMethod.DEFAULT
    ]
    task_request_msg_gen = default_authoriser.create_task_request_message_generator()
    return task_request_msg_gen


@fixture
def saml_task_request_msg_gen() -> _TaskRequestMessageGenerator:
    """Creates a TaskRequestMessageGenerator for the SAML verification method."""
    # Create SAML task_request_msg_gen
    saml_authoriser = _IDENTITY_VERIFICATION_METHODS_MAP[
        IdentityVerificationMethod.SAML
    ]
    task_request_msg_gen = saml_authoriser.create_task_request_message_generator()
    return task_request_msg_gen


@dataclass
class KeyBasedGenerator:
    """Container for the key and generator for key-based task_request_msg_gen."""

    key: NonCallableMock
    gen: _TaskRequestMessageGenerator


@fixture
def key_based_task_request_msg_gen() -> KeyBasedGenerator:
    """Creates a TaskRequestMessageGenerator for the key-based verification method."""
    # Create key-based task_request_msg_gen
    key_based_authoriser: _SignatureBasedAuthorisation = cast(
        _SignatureBasedAuthorisation,
        _IDENTITY_VERIFICATION_METHODS_MAP[IdentityVerificationMethod.KEYS],
    )

    private_key: NonCallableMock = create_autospec(RSAPrivateKey, instance=True)
    task_request_msg_gen = key_based_authoriser.create_task_request_message_generator(
        private_key
    )

    return KeyBasedGenerator(private_key, task_request_msg_gen)


@fixture
def mock_task_request_msg_gen() -> Mock:
    """Creates a mock TaskRequestMessageGenerator."""
    # Create mock task_request_msg_gen
    task_request_msg_gen: Mock = create_autospec(_TaskRequestMessageGenerator)

    return task_request_msg_gen
